/*
 * Copyright (c) 1999 - 2001, Artur Merke <amerke@ira.uka.de> 
 *
 * This file is part of FrameView2d.
 *
 * FrameView2d is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * FrameView2d is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FrameView2d; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "ascii_dev.h"
#include "ascii_processor.h"

AsciiDevice::AsciiDevice(int port) {
  if (use_socket) {
    sock.init_socket_fd(port);
    sock.set_fd_nonblock();
    //sock.init_serv_addr("host",port); // nur als Empfaenger von Nachrichten konzipiert
  }

  state_string[0]= '\0';
}

bool AsciiDevice::init_frames(BuilderBase * build) {
  return AsciiProcessor::scan_and_parse(build,"
VA (0,1.5,4,4)
INS FRAME id=1 
INS 1 LINE (0,0,0,1)
INS 1 FRAME id= 2 (0,1,0.78)
INS 2 LINE col=ff0000 (0,0,0,1)
INS 1 FRAME id= 3 (0,1)
INS 3 LINE col=00ff00 (0,0,0,1)
INS 1 FRAME id= 4 (0,1,-0.78)
INS 4 LINE col=0000ff (0,0,0,1)
INS 2 FRAME id= 5 (0,1,0.52)
INS 5 LINE col=ffcc00 (0,0,0,1)
INS 2 FRAME id= 6 (0,1,-0.52)
INS 6 LINE col=ff00cc (0,0,0,1)
INS 3 FRAME id= 7 (0,1,0.52)
INS 7 LINE col=ccff00 (0,0,0,1)
INS 3 FRAME id= 8 (0,1,-0.52)
INS 8 LINE col=00ffcc (0,0,0,1)
INS 4 FRAME id= 9 (0,1,0.52)
INS 9 LINE col=cc00ff (0,0,0,1)
INS 4 FRAME id= 10 (0,1,-0.52)
INS 10 LINE col=00ccff (0,0,0,1)");
#ifdef BLABLA
  //both versions produce the same code!
#if 1
  return AsciiProcessor::scan_and_parse(build,"
INS FRAME id= 1 
INS 1 POLYGON fil=1 col= cc3333 (-10,10) (10,10) (10,-10) (-10,-10)
INS 1 POLYLINE col= 000000 lay= 1 (0,0) (20,0) (30,30)");
#else
  RGBcolor c_car(204,51,51);
  RGBcolor c_pole(0,0,0);

  const int num_points= 10;
  
  Multi<Point2d> points;
  points.set_max_size(num_points);

  int cur_frame= 1;
  build->set_cmd_insert_frame(0,cur_frame,Point2d(0.0,0.0),0.0,1);

  //field
  int k= 0;
  points.tab[k++]= Point2d(-10, 10);
  points.tab[k++]= Point2d( 10, 10);
  points.tab[k++]= Point2d( 10, -10);
  points.tab[k++]= Point2d(-10, -10);
  points.set_cur_size(k);
  build->set_cmd_insert_f_polygon(cur_frame,0,points,0, c_car);

  k= 0;
  points.tab[k++]= Point2d(0, 0);
  points.tab[k++]= Point2d(20,0);
  points.tab[k++]= Point2d(30,30);

  points.set_cur_size(k);
  build->set_cmd_insert_polyline(cur_frame,0,points,1, c_pole);

  return true;
#endif
#endif
}

bool AsciiDevice::process_input(fd_set *set, BuilderBase * build) {
  //EnDeCoderBin enc;
  static char  buffer[2][BUFFER_MAX_SIZE];
  int buffer_idx= 0;
  char *dum= buffer[buffer_idx];
  int num_msg= 0;
  int num_bytes= 0;
  bool got_input= false;

  if (use_socket && is_fd_in_set(sock.socket_fd, set) ) {
    //cout << "\n----------------------------------------";
    int num_msg_squared= 0;
    while ( sock.recv_msg( dum, num_bytes, BUFFER_MAX_SIZE, true) ) {
      //bool processed= false;
      dum[num_bytes]= '\0';
      
      while ( *dum== ' ')
	dum++;
      if ( *dum=='[' ) {
	while ( *dum != '\0' && *dum != ']' )
	  dum++;

	if ( * dum == ']' ) {
	  dum++;
	  AsciiProcessor::scan_and_parse(build,dum);
	  cout << "\n msg num= " << num_msg_squared++ << ", bytes=" << num_bytes << flush;
	  got_input= true;
	}
	else {
	  cerr << "ascii_dev: no valid protocol option, skipping message";
	}
	continue;
      }

      buffer_idx= (buffer_idx+1)%2;
      dum= buffer[buffer_idx];
      num_msg+= 1;
	//cout << "\nnum_bytes= " << num_bytes;
      
    }

    buffer_idx= (buffer_idx+1)%2;
    dum= buffer[buffer_idx];

    if ( num_msg ) {
      if (num_msg > 1)
      cout << "\nSKIPPED " << num_msg-1 << " MESSAGE(S)" << flush;
      //cout << "\nbuf: " << dum << flush;
      AsciiProcessor::scan_and_parse(build,dum);
      got_input= true;
    }
  }

  if ( use_stdin && is_fd_in_set(0, set) ) {
    cout << "\n[got input]";
    cin.getline(buffer[0],BUFFER_MAX_SIZE);
    AsciiProcessor::scan_and_parse(build,buffer[0]);
    got_input= true; 
  }

  return got_input;
}
#if 1
int  AsciiDevice::set_fds(fd_set * set) { 
  int max= 0;
  if ( use_socket ) {
    add_fd_to_set(sock.socket_fd,set); 
    max= sock.socket_fd;
  }
  if ( use_stdin )
    add_fd_to_set(0,set); //standard input
  return sock.socket_fd;
}

bool AsciiDevice::got_fds(fd_set * set) { 
  if (use_socket)
    if ( is_fd_in_set(sock.socket_fd,set) ) return true;
  if (use_stdin)
    return is_fd_in_set(0,set);
  return false;
}
#endif
