/*
 * Copyright (c) 1999 - 2001, Artur Merke <amerke@ira.uka.de> 
 *
 * This file is part of FrameView2d.
 *
 * FrameView2d is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * FrameView2d is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FrameView2d; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "ascii_processor.h"
#include "strconv.h"
#include <stdlib.h>
#include <stdio.h>

#define P_ERROR(xxx) cerr << "\nparse error: " << xxx
#define P_WARNING(xxx) cerr << "\nparse warning: " << xxx
#define P_MSG(xxx) //cerr << "\nparse msg: " << xxx

Multi<Point2d>     AsciiProcessor::points(20);
Multi<Line2d>      AsciiProcessor::lines(10);
Multi<Circle2d>    AsciiProcessor::circles(10);
Multi<CircleArc2d> AsciiProcessor::circlearcs(10);
Multi<char>        AsciiProcessor::characters(100);


bool AsciiProcessor::
mov_cmd(const char * buf, BuilderBase * build, const char* & next) {
  const char * dum;
  int fref;
  Point2d pos;
  Angle ang;
  if ( ! substrconv(buf,fref,dum) ) {
    P_ERROR("MOV no frame reference specified");
    return CMD_ERR;
  }
  buf= dum;
	
  int res= get_frame(buf,pos,ang,dum);
	
  if ( res < 1) {
    P_ERROR("MOV no frame data specified");
    return CMD_ERR;
  }
  buf= dum;
  switch (res) {
  case 3:
    build->set_cmd_set_frame_pos_ang(fref,pos,ang);
    P_MSG("build->set_cmd_set_frame_pos_ang(fref,("<<pos.x<<","<<pos.y<<"),"<< ang.get_value() <<");");
    break;
  case 2:
    build->set_cmd_set_frame_pos(fref,pos);
    P_MSG("build->set_cmd_set_frame_pos(fref,pos)");
    break;
  case 1:
    build->set_cmd_set_frame_ang(fref,ang);
    P_MSG("build->set_cmd_set_frame_ang(fref,ang);");
    break;
  }
  next= buf;
  return CMD_OK;
}

bool AsciiProcessor::
ins_cmd(const char * buf, BuilderBase * build,const char* & next) {
  const char * dum;

  int fref= 0;
  if ( strconv(buf,fref,dum) ) 
    buf= dum;
  int id= 0;
  int lay= 0;
  int vis= 1;
  RGBcolor col(0,0,0);
  int fil= 0;
  // ==========================================================
  if ( strskip(buf,"FRAME",dum) ) {
    buf= dum;
    Point2d pos;
    Angle ang;
    if ( ! get_frame_attr(buf,id,lay,vis, dum) ) {
      P_ERROR("wrong frame attributes");
      return CMD_ERR;
    }
    buf= dum;
    if ( get_frame(buf, pos, ang, dum) < -1 ) {
      P_ERROR("wrong frame data");
      return CMD_ERR;
    }
    buf= dum;
    build->set_cmd_insert_frame(fref,id,pos,ang,lay);
    P_MSG("build->set_cmd_insert_frame(" << fref << "," << id << "," << "pos,ang" << "," << lay << ");");
  } 
  // ==========================================================
  else if ( strskip(buf,"CIRCLE",dum) ) {
    buf= dum;
    if ( ! get_obj_attr(buf,id,lay,col,fil,dum) ) {
      P_ERROR("INS wrong object arguments");
      return CMD_ERR;
    }
	    
    buf= dum;
    //cout << "\nGOT CIRCLE id=" << id << " lay=" << lay << flush;

    int num= get_circles(buf,circles,dum );
    buf= dum;
    if ( num < 1 ) {
      P_ERROR("CIRCLE wrong circle(s) format");
      return CMD_ERR;
    }
    //cout << "\n//////  num_circles= " << num;
    if (num==1) {
      if (!fil) {
	build->set_cmd_insert_circle(fref,id,circles.tab[0],lay,col);
	P_MSG("build->set_cmd_insert_circle(fref,id,circles.tab[0],lay,col);");
      }
      else {
	build->set_cmd_insert_f_circle(fref,id,circles.tab[0],lay,col);
	P_MSG("build->set_cmd_insert_f_circle(fref,id,circles.tab[0],lay,col);");
      }
    }
    else {
      if (!fil) {
	build->set_cmd_insert_circles(fref,id,circles,lay,col);
	P_MSG("build->set_cmd_insert_circles(" << fref << "," << id << ",circles[num="<< circles.cur_size <<"]," << lay <<",col);");
      }
      else {
	build->set_cmd_insert_f_circles(fref,id,circles,lay,col);
	P_MSG("build->set_cmd_insert_f_circles(" << fref << "," << id << ",circles[num="<< circles.cur_size <<"]," << lay <<",col);");
      }
    }
  }
  // ==========================================================
  else if ( strskip(buf,"LINE",dum) ) {
    buf= dum;
    if ( ! get_obj_attr(buf,id,lay,col,fil,dum) ) {
      P_ERROR("INS wrong object arguments");
      return CMD_ERR;
    }
	    
    buf= dum;
    //cout << "\nGOT LINE id=" << id << " lay=" << lay << flush;

    int num= get_lines(buf,lines,dum );
    buf= dum;
    if ( num < 1 ) {
      P_ERROR("LINE wrong line(s) format");
      return CMD_ERR;
    }
    if (num==1) {
      build->set_cmd_insert_line(fref,id,lines.tab[0],lay,col);
      P_MSG("build->set_cmd_insert_line(fref,id,lines.tab[0],lay,col);");
    }
    else {
      build->set_cmd_insert_lines(fref,id,lines,lay,col);
      P_MSG("build->set_cmd_insert_lines(" << fref << "," << id << ",lines," << lay <<",col);");
    }
  }
  // ==========================================================
  else if ( strskip(buf,"POINT",dum) ) {
    buf= dum;
    if ( ! get_obj_attr(buf,id,lay,col,fil,dum) ) {
      P_ERROR("INS wrong object arguments");
      return CMD_ERR;
    }
	    
    buf= dum;
    //cout << "\nGOT POINT id=" << id << " lay=" << lay << flush;

    int num= get_points(buf,points,dum );
    buf= dum;
    if ( num < 1 ) {
      P_ERROR("POINT wrong point(s) format");
      return CMD_ERR;
    }
    if (num==1) {
      build->set_cmd_insert_point(fref,id,points.tab[0],lay,col);
      P_MSG("build->set_cmd_insert_point(fref,id,points.tab[0],lay,col);");
    }
    else {
      build->set_cmd_insert_points(fref,id,points,lay,col);
      P_MSG("build->set_cmd_insert_points(" << fref << "," << id << ",points," << lay <<",col);");
    }
  }
  // ==========================================================
  else if ( strskip(buf,"POLYLINE",dum) ) {
    buf= dum;
    if ( ! get_obj_attr(buf,id,lay,col,fil,dum) ) {
      P_ERROR("POLYLINE wrong object arguments");
      return CMD_ERR;
    }
	    
    buf= dum;
    //cout << "\nGOT POINT id=" << id << " lay=" << lay << flush;

    int num= get_points(buf,points,dum );
    buf= dum;
    if ( num < 2 ) {
      P_ERROR("POLYLINE wrong point(s) format");
      return CMD_ERR;
    }
    build->set_cmd_insert_polyline(fref,id,points,lay,col);
    P_MSG("build->set_cmd_insert_polyline(" << fref << "," << id << ",points," << lay <<",col);");
  }
  // ==========================================================
  else if ( strskip(buf,"POLYGON",dum) ) {
    buf= dum;
    if ( ! get_obj_attr(buf,id,lay,col,fil,dum) ) {
      P_ERROR("INS wrong object arguments");
      return CMD_ERR;
    }
	    
    buf= dum;
    //cout << "\nGOT POINT id=" << id << " lay=" << lay << flush;

    int num= get_points(buf,points,dum );
    buf= dum;
    if ( num < 2 ) {
      P_ERROR("POLYGON wrong point(s) format");
      return CMD_ERR;
    }
    if (!fil) {
      build->set_cmd_insert_polygon(fref,id,points,lay,col);
      P_MSG("build->set_cmd_insert_polygon(" << fref << "," << id << ",points," << lay <<",col);");
    }
    else {
      build->set_cmd_insert_f_polygon(fref,id,points,lay,col);
      P_MSG("build->set_cmd_insert_f_polygon(" << fref << "," << id << ",points," << lay <<",col);");
    }
  }
  // ==========================================================
  else if ( strskip(buf,"STRING",dum) ) {
    Point2d point;
    buf= dum;
    if ( ! get_obj_attr(buf,id,lay,col,fil,dum) ) {
      P_ERROR("STRING wrong object arguments");
      return CMD_ERR;
    }
    buf= dum; 
    //format ( d0, d1, "<string>" )
    if ( !strshift(buf,"(",dum) )
      return CMD_ERR;
    buf= dum;
    if ( !substrconv(buf,point.x,dum) ) 
      return CMD_ERR; 
    buf= dum;
    if ( !strshift(buf,",",dum) )
      return CMD_ERR;
    buf= dum;
    if ( !substrconv(buf,point.y,dum) ) 
      return CMD_ERR; 
    buf= dum;
    if ( !strshift(buf,",",dum) )
      return CMD_ERR;
    buf= dum;

    int offset;
    int res= get_string(buf,offset,dum);
    if (res < 0) {
      P_ERROR("INS STRING  wrong text specification, retval= " << res);
      return CMD_ERR;
    }
    VisualObject2d * string2d= new VisualString2d(id,lay,col,point,res,buf+offset);
    buf= dum;
    if ( !strshift(buf,")",dum) )
      return CMD_ERR;
    buf= dum;

    P_MSG("build->set_cmd_insert_visobject(" << fref << ",string2d)");
    if (!build->set_cmd_insert_visobject(fref,string2d)) {
      P_ERROR("INS STRING problem with inserting STRING");
      delete string2d;
    }
  }
  else {
    P_ERROR("INS: no such element");
    return CMD_ERR;
  }
  next= buf;
  return CMD_OK;
}

bool AsciiProcessor::
visual_area_cmd(const char * buf, BuilderBase * build,const char* & next) {
  const char * dum;
  Area2d a;
  double d1;
  int res= get_max5_tuple(buf,4,a.center.x,a.center.y,a.size_x,a.size_y,d1,dum);
  if (res != 4) {
    P_ERROR("SETAREA wrong area specification, retval= " << res);
    return CMD_ERR;
  }
  buf= dum;
  P_MSG("build->set_cmd_set_view_area(area[("<< a.center.x << "," << a.center.y << ")," << a.size_x << "," << a.size_y << "]" <<   ");");
  build->set_cmd_set_view_area(a);
  next= dum;
  return CMD_OK;
}

bool AsciiProcessor::
status_line_cmd(const char * buf, BuilderBase * build,const char* & next) {
  const char * dum;
  int res= get_string(buf,characters,dum);
  if (res < 0) {
    P_ERROR("SL wrong text specification, retval= " << res);
    return CMD_ERR;
  }
  buf= dum;
  P_MSG("build->set_cmd_set_status(characters);");
  build->set_cmd_set_status_line(characters);
  next= dum;
  return CMD_OK;
}

bool AsciiProcessor::
background_color_cmd(const char * buf, BuilderBase * build,const char* & next) {
  const char * dum;
  RGBcolor col;

  if ( !get_col(buf,col,dum) ) {
    P_ERROR("wrong color [" << buf << "]");
    return false;
  }
  build->set_cmd_set_background_color(col);
  next= dum;
  return CMD_OK;
}

bool AsciiProcessor::
del_cmd(const char * buf, BuilderBase * build,const char* & next) {
  const char * dum;
  int fref;
  if ( !strconv(buf,fref,dum) ) {
    P_ERROR("DEL no frame or object reference");
    return CMD_ERR;
  }
  buf= dum;
  int oref;
  if ( strconv(buf,oref,dum) ) {
    buf= dum;
    build->set_cmd_remove_object(fref,oref);
    P_MSG("build->set_cmd_remove_object(fref,oref);");
  }
  else {
    build->set_cmd_remove_frame(fref);
    P_MSG("build->set_cmd_remove_frame(fref);");
  }
  next= buf;
  return CMD_OK;
}

bool AsciiProcessor::
emp_cmd(const char * buf, BuilderBase * build,const char* & next) {
  const char * dum;
  int fref;
  if ( !strconv(buf,fref,dum) ) {
    P_ERROR("EMP no frame reference");
    return CMD_ERR;
  }
  buf= dum;
  build->set_cmd_empty_frame(fref);
  P_MSG("build->set_cmd_empty_frame(" << fref << ");");
  next= buf;
  return CMD_OK;
}

#if 0
      if ( strskip(buf,"SETVIS",dum) ) {
	buf= dum;
	int fref;
	if ( strconv(buf,fref,dum) ) {
	  buf= dum;
	  int oref;
	  if ( strconv(buf,oref,dum) ) {
	    buf= dum;
	    build->set_cmd_set_object_visible(fref,oref,1);
	    P_MSG("build->set_cmd_set_object_visible(fref,oref,1);");
	  }
	  else {
	    build->set_cmd_set_frame_visible(fref,1);
	    P_MSG("build->set_cmd_set_frame_visible(fref,1);");
	  }
	}
	else 
	  P_ERROR("SETVIS no frame or object reference");
      } //+++++++++++++++++++++++++++++++++++++++
      else if ( strskip(buf,"SETCOL",dum) ) {
      }
#endif

bool AsciiProcessor::scan_and_parse(BuilderBase * build, const char * buf) {
  while (buf) {
    while ( *buf==' ' || *buf=='\n' || *buf=='\t' )
      buf++;
    if ( *buf == '\0')
      break;

    const char *dum;
    bool res= false;
    switch (buf[0]) {
    //------------------------------------------------------------------------
    case 'B':
      if ( buf[1] == 'G') 
	if ( background_color_cmd(buf+2, build, dum) ) {
	  res= true;
	  buf= dum;
	  //cout << "col set" << flush;
	}
      break;
    //------------------------------------------------------------------------
    case 'M':
      if ( buf[1] == 'O' && buf[2] == 'V' ) 
	if ( mov_cmd(buf+3, build, dum) ) {
	  res= true;
	  buf= dum;
	}
      break;
    //------------------------------------------------------------------------
    case  'S':
      if ( buf[1]== 'L')
	if ( status_line_cmd(buf+2,build,dum) ) {
	  res= true;
	  buf= dum;
	}
      break;
    //------------------------------------------------------------------------
    case  'D':
      if ( buf[1] == 'E' && buf[2]== 'L' )
	if ( del_cmd(buf+3, build, dum) ) {
	  res= true;
	  buf= dum;
	}
      break;
    //------------------------------------------------------------------------
    case  'E':
      if ( buf[1] == 'M' && buf[2]== 'P' )
	if ( emp_cmd(buf+3, build, dum) ) {
	  res= true;
	  buf= dum;
	}
      break;
    //------------------------------------------------------------------------
    case 'I':
      if ( buf[1] == 'N' && buf[2]== 'S' )
	if ( ins_cmd(buf+3, build, dum) ) {
	  res= true;
	  buf= dum;
	}
      break;
    //------------------------------------------------------------------------
    case  'V':
      if ( buf[1] == 'A' )
	if ( visual_area_cmd(buf+2, build, dum) ) {
	  res= true;
	  buf= dum;
	}
      break;
    //------------------------------------------------------------------------
    }
    if (!res) {
      P_ERROR("unrecognized command [" << buf << "]");
      break;
    }
    //P_MSG("buf [" << buf << "]");
  }
  return true;
}

bool AsciiProcessor:: get_obj_attr(const char * buf, int & id, int & lay, RGBcolor & col, int & fil,char const* & next) {
  const char *dum= buf;
  bool got_attr[4]= { false, false, false, false };
  while (dum) {
    if ( strshift(buf,"id",dum) ) {
      if ( got_attr[0] )
	return false;
      
      got_attr[0]= true;
      buf= dum;
      if (!strshift(buf,"=",dum) )
	return false;

      buf= dum;
      if ( !strconv(buf,id,dum) )
	return false;
      buf= dum;
    }
    else if ( strshift(buf,"lay",dum) ) {
      if ( got_attr[1] )
	return false;
      
      got_attr[1]= true;
      buf= dum;
      if (!strshift(buf,"=",dum) )
	return false;

      buf= dum;
      if ( !strconv(buf,lay,dum) )
	return false;
      buf= dum;
    }
    else if ( strshift(buf,"col",dum) ) {
      if ( got_attr[2] )
	return false;
      
      got_attr[2]= true;
      buf= dum;
      if (!strshift(buf,"=",dum) )
	return false;

#if 0
      buf= dum;
      if (!strshift(buf,"#",dum) )
	return false;
#endif
      buf= dum;
      if ( !get_col(buf,col,dum) ) {
	P_ERROR("wrong color [" << buf << "]");
	return false;
      }
      buf= dum;
      //if ( !strconv(buf,col,dum) )
      //return false;
      buf= dum;
    }
    else if ( strshift(buf,"fil",dum) ) {
      if ( got_attr[3] )
	return false;
      
      got_attr[1]= true;
      buf= dum;
      if (!strshift(buf,"=",dum) )
	return false;

      buf= dum;
      bool b;
      if ( !strconv(buf,b,dum) )
	return false;
      fil= b;
      buf= dum;
    }
    else { 
      next= buf;
      return true;
    }
  }
  next= buf;
  return true;
}

bool AsciiProcessor:: get_frame_attr(const char * buf, int & id, int & lay, int & vis, char const* & next) {
  const char *dum= buf;
  bool got_attr[4]= { false, false, false, false };
  while (dum) {
    if ( strshift(buf,"id",dum) ) {
      if ( got_attr[0] )
	return false;
      
      got_attr[0]= true;
      buf= dum;
      if (!strshift(buf,"=",dum) )
	return false;

      buf= dum;
      if ( !strconv(buf,id,dum) )
	return false;
      buf= dum;
    }
    else if ( strshift(buf,"lay",dum) ) {
      if ( got_attr[1] )
	return false;
      
      got_attr[1]= true;
      buf= dum;
      if (!strshift(buf,"=",dum) )
	return false;

      buf= dum;
      if ( !strconv(buf,lay,dum) )
	return false;
      buf= dum;
    }
    else if ( strshift(buf,"vis",dum) ) {
      bool b;
      if ( got_attr[1] )
	return false;
      
      got_attr[1]= true;
      buf= dum;
      if (!strshift(buf,"=",dum) )
	return false;

      buf= dum;
      if ( !strconv(buf,b,dum) )
	return false;
      vis= b;
      buf= dum;
    }
    else { 
      next= buf;
      return true;
    }
  }
  next= buf;
  return true;
}

bool AsciiProcessor::get_col(const char * buf, RGBcolor & col, const char* & next) {
  while (*buf == ' ')
    buf++;
  if ( buf == '\0' )
    return false;

  unsigned char c[3];
  bool res= false;
  for (int i=0; i<3; i++) {
    if ( buf[0]!= '\0' && buf[1] != '\0') {
      char chr[3];
      chr[0]= buf[0]; chr[1]= buf[1]; chr[2]= '\0';
      char* error_ptr;
      long tmp_long= strtol(chr, &error_ptr, 16);
      if (error_ptr!= chr+2 || tmp_long < 0 || tmp_long > 255) {
	cerr << "\ntmp_long=" << tmp_long;
	break;
      }
      c[i]= (unsigned char) tmp_long;
      buf+= 2;
      if (i==2)
	res= true;
    }
    else
      break;
  }
  if (res== true) {
    next= buf;
    col= RGBcolor(c[0],c[1],c[2]);
    return true;
  }
  return false;
}
/* results:
   0 : a empty tuple was received "()" 
   1 : a one tuple "( <num> )"
   2,3,4,5 analogous with 1

   -1: the was no tuple at all, i.e. the string doesn't begin with a "("
   -2: some failure occured
*/

int AsciiProcessor::get_max5_tuple(const char *buf, int num, double & d0, double & d1, double & d2, double & d3, double & d4, char const* & next) {
  const char * dum;
  if ( !strshift(buf,"(",dum) )
    return -1;
  buf= dum;
  if ( !substrconv(buf,d0,dum) ) {
    if ( strshift(buf,")",dum) ) {
      next= dum;
      return 0; 
    }
    return -2;
  }
  if (num<1) return -2;
    
  buf=dum;
  if ( !strshift(buf,",",dum) ) {
    if ( strshift(buf,")",dum) ) {
      next= dum;
      return 1;
    }
    return -2;
  }
  if (num<2) return -2;

  buf= dum;
  if ( !substrconv(buf,d1,dum) )
    return -2;
  buf= dum;
  if ( !strshift(buf,",",dum) ) {
    if (strshift(buf,")",dum) ) {
      next= dum;
      return 2;
    }
    return -2;
  }
  if (num<3) return -2;

  buf= dum;
  if ( !substrconv(buf,d2,dum) )
    return -2;
  buf= dum;
  if ( !strshift(buf,",",dum) ) {
    if (strshift(buf,")",dum) ) {
      next= dum;
      return 3;
    }
    return -2;
  }
  if (num<4) return -2;

  buf= dum;
  if ( !substrconv(buf,d3,dum) )
    return -2;
  buf= dum;
  if ( !strshift(buf,",",dum) ) {
    if (strshift(buf,")",dum) ) {
      next= dum;
      return 4;
    }
    return -2;
  }
  if (num<5) return -2;

  buf= dum;
  if ( !substrconv(buf,d4,dum) )
    return -2;
  buf= dum;
  if ( !strshift(buf,")",dum) ) 
    return -2;

  next= dum;
  return 5;
}

int AsciiProcessor::get_frame(const char * buf, Point2d & p, Angle & a, char const* & next) {
  double d0,d1,d2,d3,d4;
  int res= get_max5_tuple(buf,3,d0,d1,d2,d3,d4,next);
  if (res>1) {
    p.x= d0; p.y= d1;
    if (res==3)
      a= Angle(d2);
  }
  else 
    if (res==1) 
      a= Angle(d0);

  return res;
}

int AsciiProcessor::get_point(const char * buf, Point2d & p, char const* & next) {
  double d2,d3,d4;
  return get_max5_tuple(buf,2,p.x,p.y,d2,d3,d4,next);
}

int AsciiProcessor::get_line(const char * buf, Line2d & l, char const* & next) {
  double d4;
  return get_max5_tuple(buf,4,l.p1.x,l.p1.y,l.p2.x,l.p2.y,d4,next);
}

int AsciiProcessor::get_circle(const char * buf, Circle2d & c, char const* & next) {
  double d3,d4;
  return get_max5_tuple(buf,3,c.center.x,c.center.y,c.radius,d3,d4,next);
}

int AsciiProcessor::get_circlearc(const char * buf, CircleArc2d & c, char const* & next) {
  double d3,d4;
  int res= get_max5_tuple(buf,5,c.center.x,c.center.y,c.radius,d3,d4,next);
  if (res>3)
    c.ang1= Angle(d3);
  if (res>4)
    c.ang2= Angle(d4);
  return res;
}

int AsciiProcessor::get_points(const char *buf, Multi<Point2d> & points, char const* & next) {
  int res;
  int k=0;
  const char * dum;
  while (1) {
    if ( k>= points.max_size ) {
      points.set_cur_size(k); //do not loose the points
      points.set_max_size_and_preserve_cur_size(k+k+1);
    }

    res= get_point(buf,points.tab[k],dum);

    if (res == -1)
      break;

    if (res != 2)
      return -1;

    buf= dum;
    k++;
  };

  next= dum;

  points.set_cur_size(k);
  return k;
}

int AsciiProcessor::get_lines(const char *buf, Multi<Line2d> & lines, char const* & next) {
  int res;
  int k=0;
  const char * dum;
  while (1) {
    if ( k>= lines.max_size ) {
      lines.set_cur_size(k); //do not loose entries
      lines.set_max_size_and_preserve_cur_size(k+k+1);
    }
    res= get_line(buf,lines.tab[k],dum);

    if (res == -1)
      break;

    if (res != 4)
      return -1;

    buf= dum;
    k++;
  };

  next= dum;

  lines.set_cur_size(k);
  return k;
}

int AsciiProcessor::get_circles(const char *buf, Multi<Circle2d> & circles, char const* & next) {
  int res;
  int k=0;
  const char * dum;
  while (1) {
    if ( k>= circles.max_size ) {
      circles.set_cur_size(k); //do not loose entries
      circles.set_max_size_and_preserve_cur_size(k+k+1);
    }
    res= get_circle(buf,circles.tab[k],dum);

    if (res == -1)
      break;

    if (res != 3)
      return -1;

    buf= dum;
    k++;
  };

  next= dum;

  circles.set_cur_size(k);
  return k;
}

int AsciiProcessor::get_circlearcs(const char *buf, Multi<CircleArc2d> & circles, char const* & next) {
  int res;
  int k=0;
  const char * dum;
  while (1) {
    if ( k>= circles.max_size ) {
      circles.set_cur_size(k); //do not loose entries
      circles.set_max_size_and_preserve_cur_size(k+k+1);
    }

    res= get_circlearc(buf,circles.tab[k],dum);

    if (res == -1)
      break;

    if (res != 5 && res != 3)
      return -1;

    buf= dum;
    k++;
  };

  next= dum;

  circles.set_cur_size(k);
  return k;
}

int AsciiProcessor::get_string(const char * buf, Multi<char> & str,  char const* & next) {
  const char * dum;
  if ( !strshift(buf,"\"",dum) )
    return -1;
  buf= dum;
  int k=0;
  while ( *buf !=  '\"' && *buf != '\0') {
    if ( k>= str.max_size ) {
      str.set_cur_size(k); //do not loose entries
      str.set_max_size_and_preserve_cur_size(k+k+1);
    }
    str.tab[k]= *buf;
    buf++;
    k++;
  }
  if ( *buf != '\"' )
    return -2;
  buf++;
  next= buf;
  str.set_cur_size(k);
  return k;
}

int AsciiProcessor::get_string(const char * buf, int & offset,  char const* & next) {
  const char * dum;
  if ( !strshift(buf,"\"",dum) )
    return -1;

  offset= dum-buf;
  buf= dum;

  int k=0;
  while ( *buf !=  '\"' && *buf != '\0') {
    buf++;
    k++;
  }
  if ( *buf != '\"' )
    return -2;
  buf++;
  next= buf;
  return k;
}
