/*
 * Copyright (c) 1999 - 2001, Artur Merke <amerke@ira.uka.de>
 *
 * This file is part of FrameView2d.
 *
 * FrameView2d is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * FrameView2d is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FrameView2d; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "conv.h"

#include <math.h>
#include <iostream.h>

#define _FNAME_ "\n(view.C):"


ostream& operator<< (ostream& o, const WinPlaneConverter& c) {
  cout << "[ Win= ("<< c.Win_width << ", " << c.Win_height << ")"
       << "  Plane= (" << c.Plane_east_x << ", " << c.Plane_south_y << ")"
       << "  (" << c.Plane_west_x << ", " << c.Plane_north_y << ") ]";
  return o;
}


WinPlaneConverter::WinPlaneConverter() {
}

WinPlaneConverter::WinPlaneConverter(int width, int height, 
				     const Point2d &center, Value size_x, Value size_y) {
  init(width,height, center,size_x, size_y);
}

void WinPlaneConverter::init(int width, int height, 
			     const Point2d &center, Value size_x, Value size_y) {
  init_win(width,height);
  init_plane(center,size_x,size_y);
}

void WinPlaneConverter::init_win(int width, int height) {
  if (width < 20 || height < 20)
    cerr << _FNAME_ << "wrong view parameter";
  Win_width    = width;
  Win_height    = height;
}


void WinPlaneConverter::init_plane(const Point2d &center,Value size_x,Value size_y) {
  if (size_x < 0 || size_y < 0 )
    cerr << _FNAME_ << "wrong view parameter";
  Value hsx= size_x*0.5;
  Value hsy= size_y*0.5;
  Plane_east_x  = center.x - hsx;
  Plane_west_x  = center.x + hsx;
  Plane_north_y = center.y + hsy;
  Plane_south_y = center.y - hsy; 
}

Value WinPlaneConverter::x_Win_to_x_Plane(int x) const {
  return Plane_east_x + (Plane_west_x-Plane_east_x)*x/Win_width;
}
  
Value WinPlaneConverter::y_Win_to_y_Plane(int y) const {
  return Plane_south_y + (Plane_north_y-Plane_south_y)*(Win_height-y)/Win_height;
}
 
Value WinPlaneConverter::size_x_Win_to_size_x_Plane(int sx) const {
  return ((Plane_west_x - Plane_east_x)*sx)/ Value(Win_width);
}

Value WinPlaneConverter::size_y_Win_to_size_y_Plane(int sy) const {
  return ((Plane_north_y - Plane_south_y)*sy)/ Value(Win_height);
} 

int WinPlaneConverter::x_Plane_to_x_Win(Value x) const {
  //return (int)rint(Win_width*(x- Plane_east_x)/(Plane_west_x-Plane_east_x));
  return (int)floor(Win_width*(x- Plane_east_x)/(Plane_west_x-Plane_east_x));
}
 
int WinPlaneConverter::y_Plane_to_y_Win(Value y) const {
  //return Win_height- (int)rint(Win_height*(y-Plane_south_y)/(Plane_north_y-Plane_south_y));
  return Win_height- (int)floor(Win_height*(y-Plane_south_y)/(Plane_north_y-Plane_south_y));
}

int WinPlaneConverter::size_x_Plane_to_size_x_Win(Value s) const {
  return (int)rint( Win_width*s/(Plane_west_x-Plane_east_x));
}

int WinPlaneConverter::size_y_Plane_to_size_y_Win(Value s) const {
  return (int)rint( Win_height*s/(Plane_north_y-Plane_south_y));
}


int WinPlaneConverter::angle_Plane_to_angle_Win(const Angle& a) const {
  return (int)rint(64*a.get_value()*180.0/M_PI);
}
 
void WinPlaneConverter::get_Win(int & width, int & height) const {
  width= Win_width;
  height= Win_height;
}

void WinPlaneConverter::get_Plane(Point2d & center, Value & size_x, Value & size_y) const {
  size_x= Plane_west_x - Plane_east_x;
  size_y= Plane_north_y - Plane_south_y;
  center.x = Plane_east_x + size_x*0.5;
  center.y = Plane_south_y + size_y*0.5;
}
/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/
/******************** T E S T ********* T E S T*******************************/

#if 0

main() {
  WinPlaneConverter v(40,40,-50,-30,50,30);

  cout << "\n" << v.x_Win_to_x_Plane(v.x_Plane_to_x_Win(0.0));
  cout << "\n" << v.x_Plane_to_x_Win(-50.0);
  cout << "\n" << v.y_Plane_to_y_Win(-30.0);
}



#endif

