/*
 * Copyright (c) 1999 - 2001, Artur Merke <amerke@ira.uka.de> 
 *
 * This code is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * It is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */
#ifndef _RWBIN_H_
#define _RWBIN_H_

#include <iostream.h>

#ifndef NATIVE_INT_FORMAT
#include <netinet/in.h>
#endif

inline void rbin(istream & in, char & val) {
  in.read(&val, sizeof(char));
}

inline void rbin(istream & in,signed char & val) {
 in.read(&val, sizeof(signed char));
}

inline void rbin(istream & in,unsigned char & val) {
 in.read(&val, sizeof(unsigned char));
}

inline void rbin(istream & in,signed int & val) {
#ifdef NATIVE_INT_FORMAT
  in.read(&val, sizeof(signed int));
#else
  unsigned long nval;
  in.read(&nval, sizeof(unsigned long));
  val = (signed int) ntohl (nval);
#endif
}

inline void rbin(istream & in,signed long & val) {
#ifdef NATIVE_INT_FORMAT
  in.read(&val, sizeof(signed long));
#else
  unsigned long nval;
  in.read(&nval, sizeof(unsigned long));
  val = (signed long) ntohl (nval);
#endif
}

inline void rbin(istream & in,signed short & val) {
#ifdef NATIVE_INT_FORMAT
  in.read(&val, sizeof(signed short));
#else
  unsigned short nval;
  in.read(&nval, sizeof(unsigned short));
  val = (signed short) ntohs (nval);
#endif
}

inline void rbin(istream & in,unsigned int & val) {
#ifdef NATIVE_INT_FORMAT
  in.read(&val, sizeof(unsigned int));
#else
  unsigned long nval;
  in.read(&nval, sizeof(unsigned long));
  val = (unsigned int) ntohl (nval);
#endif
}

inline void rbin(istream & in,unsigned long & val) {
#ifdef NATIVE_INT_FORMAT
  in.read(&val, sizeof(unsigned long));
#else
  unsigned long nval;
  in.read(&nval, sizeof(unsigned long));
  val = ntohl (nval);
#endif
}

inline void rbin(istream & in,unsigned short & val) {
#ifdef NATIVE_INT_FORMAT
  in.read(&val, sizeof(unsigned short));
#else
  unsigned short nval;
  in.read(&nval, sizeof(unsigned short));
  val = ntohs (nval);
#endif
}

inline void rbin(istream & in,bool & val) {
#ifdef NATIVE_INT_FORMAT
  in.read(&val, sizeof(bool));
#else
  unsigned char charCast;
  in.read(&charCast, sizeof(unsigned char));
  val = (bool) charCast;
#endif
}

inline void rbin(istream & in,float & val) {
  in.read(&val, sizeof(float));
}

inline void rbin(istream & in,double & val) {
  in.read(&val, sizeof(double));
}

inline void rbin(istream & in,long double & val) {
  in.read(&val, sizeof(long double));
}

/******************************************************************************/

inline void wbin(ostream & out, signed char val) {
  out.write((const void*) &val, sizeof(signed char));
}

inline void wbin(ostream & out, char val) {
  out.write((const void*) &val, sizeof(char));
}

inline void wbin(ostream & out, unsigned char val) {
  out.write((const void*) &val, sizeof(unsigned char));
}

inline void wbin(ostream & out, signed int val) {
#ifdef NATIVE_INT_FORMAT
  out.write((const void*) &val, sizeof(signed int));
#else
  unsigned long nval = htonl (val);
  out.write((const void*) &nval, sizeof(unsigned long));
#endif
}

inline void wbin(ostream & out, signed long val) {
#ifdef NATIVE_INT_FORMAT
  out.write((const void*) &val, sizeof(signed long));
#else
  unsigned long nval = htonl (val);
  out.write((const void*) &nval, sizeof(unsigned long));
#endif
}

inline void wbin(ostream & out, signed short val) {
#ifdef NATIVE_INT_FORMAT
  out.write((const void*) &val, sizeof(signed short));
#else
  unsigned short nval = htons (val);
  out.write((const void*) &nval, sizeof(unsigned short));
#endif
}

inline void wbin(ostream & out, unsigned int val) {
#ifdef NATIVE_INT_FORMAT
  out.write((const void*) &val, sizeof(unsigned int));
#else
  unsigned long nval = htonl (val);
  out.write((const void*) &nval, sizeof(unsigned long));
#endif
}

inline void wbin(ostream & out, unsigned long val) {
#ifdef NATIVE_INT_FORMAT
  out.write((const void*) &val, sizeof(unsigned long));
#else
  unsigned long nval = htonl (val);
  out.write((const void*) &nval, sizeof(unsigned long));
#endif
}

inline void wbin(ostream & out, unsigned short val) {
#ifdef NATIVE_INT_FORMAT
  out.write((const void*) &val, sizeof(unsigned short));
#else
  unsigned short nval = htons (val);
  out.write((const void*) &nval, sizeof(unsigned short));
#endif
}

inline void wbin(ostream & out, bool val) {
#ifdef NATIVE_INT_FORMAT
  out.write((const void*) &val, sizeof(bool));
#else
  unsigned char charCast = (unsigned char) val;
  out.write((const void*) &charCast, sizeof(unsigned char));
#endif
}

inline void wbin(ostream & out, float val) {
  out.write((const void*) &val, sizeof(float));
}

inline void wbin(ostream & out, double val) {
  out.write((const void*) &val, sizeof(double));
}

inline void wbin(ostream & out, long double val) {
  out.write((const void*) &val, sizeof(long double));
}

#endif 


#if 0
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/*************      T E S T     ***************************************/
#include <fstream.h>

main() {
  char c= 'C',c2;
  double d= 3.14,d2;
  int i= -254,i2;


  ofstream out("xxx.tmp");
  wbin(out,c);
  wbin(out,d);
  wbin(out,i);
  out.close();

  ifstream in("xxx.tmp");
  rbin(in,c2);
  rbin(in,d2);
  rbin(in,i2);
  in.close();

  ofstream out2("xxx2.tmp");
  wbin(out2,c2);
  wbin(out2,d2);
  wbin(out2,i2);
  out2.close();
  
  return 1;
}

#endif
