/*
 * Copyright (c) 1999 - 2001, Artur Merke <amerke@ira.uka.de> 
 *
 * This file is part of FrameView2d.
 *
 * FrameView2d is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * FrameView2d is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FrameView2d; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "smonitor_dev.h"
#include "global_defs.h"

#include <iostream>
#include <strstream>
#include <iomanip>
#include <stdio.h>			/* Standard I/O definitions. */
#include <stdlib.h>			/* Standard C Library definitions. */
#include <string.h>			/* C Library string functions. */
#include <ctype.h>  /* isprint(...) */

#include <netinet/in.h> /* wegen htons(...)*/
#include <math.h>

//#include "object2d.h"
#include "rgb_db.h"
#include "ascii_processor.h"
//#include "tools.h"

/*****************************************************************************
 \short soccerserver parameters (version 1 and 2)

 This part contains server specific data types, which are taken from the files
   - sserver-6.XX/server/param.h
   - sserver-6.XX/server/utility
 and for the extension for protocol version 2
   - sserver-7.00/server/param.h
   - sserver-7.00/server/utility
*/
namespace SSrv {
// from  utility.h
#define PI M_PI

// from param.h

#define SHOWINFO_SCALE          16.0
#define SHOWINFO_SCALE2         65536.0 //ver. 2
#define SHOWINFO_SCALE2_int     65536   //art!

#define NO_INFO         0  //ver. 2
#define SHOW_MODE       1
#define MSG_MODE        2
#define DRAW_MODE       3
#define BLANK_MODE      4
#define PM_MODE         5  //ver. 2
#define TEAM_MODE       6  //ver. 2
#define PT_MODE         7  //ver. 2
#define PARAM_MODE      8  //ver. 2
#define PPARAM_MODE     9  //ver. 2

#define DrawClear       0
#define DrawPoint       1
#define DrawCircle      2
#define DrawLine        3
#define MSG_BOARD       1
#define LOG_BOARD       2

//#define MAX_PLAYER 11

#define	DISABLE			0x0000
#define	STAND			0x0001
#define	KICK			0x0002
#define KICK_FAULT		0x0004
#define	GOALIE			0x0008
#define CATCH			0x0010
#define CATCH_FAULT		0x0020
#define BALL_TO_PLAYER	0x0040
#define PLAYER_TO_BALL	0x0080
#define DISCARD			0x0100


#define PLAYMODE_STRINGS {"",\
                        "before_kick_off",\
                        "time_over",\
                        "play_on",\
                        "kick_off_l",\
                        "kick_off_r",\
                        "kick_in_l",\
                        "kick_in_r",\
                        "free_kick_l",\
                        "free_kick_r",\
                        "corner_kick_l",\
                        "corner_kick_r",\
                        "goal_kick_l",\
                        "goal_kick_r",\
                        "goal_l",\
                        "goal_r",\
                        "drop_ball",\
                        "offside_l",\
                        "offside_r",\
                        "penalty_kick_l",\
                        "penalty_kick_r",\
                        "first_half_over",\
                        "pause",\
                        "human_judge",\
                        "foul_charge_l",\
                        "foul_charge_r",\
                        "foul_push_l",\
                        "foul_push_r",\
                        "foul_multiple_attack_l",\
                        "foul_multiple_attack_r",\
                        "foul_ballout_l",\
                        "foul_ballout_r"\
                }

#define	COLOR_NAME_MAX	64

typedef struct {
	short	enable ;
	short	side ;
	short	unum ;
	short	angle ;
	short	x ;
	short	y ;
} pos_t ;

typedef struct {
	char	name[16] ;
	short	score ;
} team_t ;

typedef struct {
	char	pmode ;
	team_t	team[2] ;
	pos_t	pos[MAX_PLAYER * 2 + 1] ;
	short	time ;
} showinfo_t ;

typedef	struct {
	short	board ;
	char	message[2048] ;
} msginfo_t ;

typedef struct {
	short	x ;
	short	y ;
	char	color[COLOR_NAME_MAX] ;
} pointinfo_t ;

typedef struct {
	short	x ;
	short	y ;
	short	r ;
	char	color[COLOR_NAME_MAX] ;
} circleinfo_t ;
	
typedef struct {
	short	x1 ;
	short	y1 ;
	short	x2 ;
	short	y2 ;
	char	color[COLOR_NAME_MAX] ;
} lineinfo_t ;

typedef struct {
	short mode ;
	union {
		pointinfo_t		pinfo ;
		circleinfo_t	cinfo ;
		lineinfo_t		linfo ;
	} object ;
} drawinfo_t ;

typedef struct {
	short	mode ;
	union {
		showinfo_t	show ;
		msginfo_t	msg ;
		drawinfo_t	draw ;
	} body ;
} dispinfo_t ;


typedef struct {
  long x;
  long y;
  long deltax;
  long deltay;
} ball_t;

typedef struct {
  short mode;
  short type;
  long x;
  long y;
  long deltax;
  long deltay;
  long body_angle;
  long head_angle;
  long view_width;
  short view_quality;
  long stamina;
  long effort;
  long recovery;
  short kick_count;
  short dash_count;
  short turn_count;
  short say_count;
  short tneck_count;
  short catch_count;
  short move_count;
  short chg_view_count;
} player_t;

typedef struct {
  char          pmode ;
  team_t        team[2] ;
  ball_t        ball;
  player_t      pos[MAX_PLAYER * 2] ;
  short         time ;
} showinfo_t2 ;  

typedef struct {
    short id;
    long player_speed_max;
    long stamina_inc_max;
    long player_decay;
    long inertia_moment;
    long dash_power_rate;
    long player_size;
    long kickable_margin;
    long kick_rand;
    long extra_stamina;
    long effort_max;
    long effort_min;

  // spare variables which are to be used for paramenter added in the future    
    long sparelong1;
    long sparelong2;
    long sparelong3;
    long sparelong4;
    long sparelong5;
    long sparelong6;
    long sparelong7;
    long sparelong8;
    long sparelong9;
    long sparelong10;
} player_type_t;

typedef struct
{
    long gwidth ;					/* goal width */
    long inertia_moment ;			/* intertia moment for turn */
    long psize ;					/* player size */
    long pdecay ;					/* player decay */
    long prand ;					/* player rand */
    long pweight ;					/* player weight */
    long pspeed_max ;				/* player speed max */
    // th 6.3.00
    long paccel_max ;				/* player acceleration max */
    //
    long stamina_max ;				/* player stamina max */
    long stamina_inc ;				/* player stamina inc */
    long recover_init ;			/* player recovery init */
    long recover_dthr ;			/* player recovery decriment threshold */
    long recover_min ;				/* player recovery min */
    long recover_dec ;				/* player recovery decriment */
    long effort_init ;				/* player dash effort init */
    long effort_dthr ;				/* player dash effort decriment threshold */
    long effort_min ;				/* player dash effrot min */
    long effort_dec ;				/* player dash effort decriment */
    long effort_ithr ;				/* player dash effort incriment threshold */
    long effort_inc ;				/* player dash effort incriment */
    // pfr 8/14/00: for RC2000 evaluation
    long kick_rand;                                /* noise added directly to kicks */
    short team_actuator_noise;                        /* flag whether to use team specific actuator noise */
    long prand_factor_l;                           /* factor to multiple prand for left team */
    long prand_factor_r;                           /* factor to multiple prand for right team */
    long kick_rand_factor_l;                       /* factor to multiple kick_rand for left team */
    long kick_rand_factor_r;                       /* factor to multiple kick_rand for right team */

    long bsize ;					/* ball size */
    long bdecay ;					/* ball decay */
    long brand ;					/* ball rand */
    long bweight ;					/* ball weight */
    long bspeed_max ;				/* ball speed max */
    // th 6.3.00
    long baccel_max;				/* ball acceleration max */
    //
    long dprate ;					/* dash power rate */
    long kprate ;					/* kick power rate */
    long kmargin ;					/* kickable margin */
    long ctlradius ;				/* control radius */
    long ctlradius_width ;			/* (control radius) - (plyaer size) */
    long maxp ;					/* max power */
    long minp ;					/* min power */
    long maxm ;					/* max moment */
    long minm ;					/* min moment */
    long maxnm ;					/* max neck moment */
    long minnm ;					/* min neck moment */
    long maxn ;					/* max neck angle */
    long minn ;					/* min neck angle */
    long visangle ;				/* visible angle */
    long visdist ;					/* visible distance */
    long windir ;					/* wind direction */
    long winforce ;				/* wind force */
    long winang ;					/* wind angle for rand */
    long winrand ;					/* wind force for force */
    long kickable_area ;			/* kickable_area */
    long catch_area_l ;			/* goalie catchable area length */
    long catch_area_w ;			/* goalie catchable area width */
    long catch_prob ;				/* goalie catchable possibility */
    short   goalie_max_moves;                 /* goalie max moves after a catch */	
    long ckmargin ;				/* corner kick margin */
    long offside_area ;			/* offside active area size */
    short win_no ;					/* wind factor is none */
    short win_random ;				/* wind factor is random */
    short say_cnt_max ;				/* max count of coach SAY */
    short SayCoachMsgSize ;				/* max length of coach SAY */
    short clang_win_size;
    short clang_define_win;
    short clang_meta_win;
    short clang_advice_win;
    short clang_info_win;
    short clang_mess_delay;
    short clang_mess_per_cycle;
    short half_time ;					/* half time */
    short sim_st ;					/* simulator step interval msec */
    short send_st ;					/* udp send step interval msec */
    short recv_st ;					/* udp recv step interval msec */
    short sb_step ;					/* sense_body interval step msec */
    short lcm_st ;		                        /* lcm of all the above steps msec */
    short SayMsgSize ;				/* string size of say message */
    short hear_max ;					/* player hear_capacity_max */
    short hear_inc ;					/* player hear_capacity_inc */
    short hear_decay ;				/* player hear_capacity_decay */
    short cban_cycle ;				/* goalie catch ban cycle */
    short slow_down_factor ;                          /* factor to slow down simulator and send intervals */
    short useoffside ;				/* flag for using off side rule */
    short kickoffoffside ;			/* flag for permit kick off offside */
    long offside_kick_margin ;		/* offside kick margin */
    long audio_dist ;				/* audio cut off distance */
    long dist_qstep ;				/* quantize step of distance */
    long land_qstep ;				/* quantize step of distance for landmark */
    long dir_qstep ;				/* quantize step of direction */
    long dist_qstep_l ;			/* team right quantize step of distance */
    long dist_qstep_r ;			/* team left quantize step of distance */
    long land_qstep_l ;			/* team right quantize step of distance for landmark */
    long land_qstep_r ;			/* team left quantize step of distance for landmark */
    long dir_qstep_l ;				/* team left quantize step of direction */
    long dir_qstep_r ;				/* team right quantize step of direction */
    short CoachMode ;				/* coach mode */
    short CwRMode ;					/* coach with referee mode */
    short old_hear ;					/* old format for hear command (coach) */
    short sv_st ;					/* online coach's look interval step */


  // spare variables which are to be used for paramenter added in the future
  long sparelong1;
  long sparelong2;
  long sparelong3;
  long sparelong4;
  long sparelong5;
  long sparelong6;
  long sparelong7;
  long sparelong8;
  long sparelong9;
  long sparelong10;

  short spareshort1;
  short spareshort2;
  short spareshort3;
  short spareshort4;
  short spareshort5;
  short spareshort6;
  short spareshort7;
  short spareshort8;
  short spareshort9;
  short spareshort10;
  
} server_params_t;

typedef struct {
    short player_types;
    short subs_max;
    short pt_max;
    
    long player_speed_max_delta_min;
    long player_speed_max_delta_max;
    long stamina_inc_max_delta_factor;
    
    long player_decay_delta_min;
    long player_decay_delta_max;
    long inertia_moment_delta_factor;

    long dash_power_rate_delta_min;
    long dash_power_rate_delta_max;
    long player_size_delta_factor;

    long kickable_margin_delta_min;
    long kickable_margin_delta_max;
    long kick_rand_delta_factor;

    long extra_stamina_delta_min;
    long extra_stamina_delta_max;
    long effort_max_delta_factor;
    long effort_min_delta_factor;

    long sparelong1;
    long sparelong2;
    long sparelong3;
    long sparelong4;
    long sparelong5;
    long sparelong6;
    long sparelong7;
    long sparelong8;
    long sparelong9;
    long sparelong10;

  short spareshort1;
  short spareshort2;
  short spareshort3;
  short spareshort4;
  short spareshort5;
  short spareshort6;
  short spareshort7;
  short spareshort8;
  short spareshort9;
  short spareshort10;

} player_params_t;

typedef struct {
  short mode;
  union {
      showinfo_t2 		show;
      msginfo_t 		msg;
      player_type_t		ptinfo;
      server_params_t		sparams;
      player_params_t		pparams;
  } body;
} dispinfo_t2 ;

}

SMonitorDevice::Options::Options() {
server_port= 6000;
  strcpy(server_host,"localhost");
  coach_port= server_port+1;
  protocol_version= 2;

  list_player_types= false;
  just_edit= false;
  track_ball= false;

  player_radius= 0.3;
  player_skin= 1;
  kick_radius= 1.085;
  ball_radius= 0.3;//0.085;
  ball_skin= 1;


  c_team_l= RGBcolor(255,255,0);
  c_goalie_l= RGBcolor(0,255,0);
  c_font_l= RGBcolor(255,0,0);//(255,0,0);
  c_invalid_l= RGBcolor(0,0,0);
  
  c_team_r= RGBcolor(0,255,255);
  c_goalie_r= RGBcolor(255,153,255);
  c_font_r= RGBcolor(0,255,0);//(255,0,0);
  c_invalid_r= RGBcolor(0,0,0);
 

  c_ball= RGBcolor(255,255,255);
  c_field= RGBcolor(0,153,0);//(153,204,51); //(204,204,204);
  c_marigin= RGBcolor(204,51,51);
  c_line= RGBcolor(255,255,255);
  c_goal= RGBcolor(0,0,0);

  info_level= info_level_def;
  scale_level= scale_level_def;
}

SMonitorDevice::ServerState::ServerState() {
  reset();
}

void SMonitorDevice::ServerState::reset() {
  //last_packet_ms_time= -1;
  reconnected= true;
  left_teamname.reset();
  right_teamname.reset();
  match_information.reset();
}

void SMonitorDevice::Positions::set_player(int i,const Point2d & p,const Angle & a) {
  player[i].pos= p;
  player[i].ang= a;
}

void SMonitorDevice::Positions::print_inline(ostream & out) {
  int prec= 4;
  out << setprecision(prec) << ball.pos.x 
      << " " << setprecision(prec) << ball.pos.y 
      << " " << setprecision(prec) << ball.vel.x 
      << " " << setprecision(prec) << ball.vel.y;
  for (int i= 0; i < MAX_PLAYER*2; i++) 
    out << "    " << setprecision(prec) << player[i].pos.x 
	<< " " << setprecision(prec) << player[i].pos.y 
	<< " " << setprecision(prec) << player[i].ang.get_value()
	<< " 0 0"; // the velocity is 0
}

void SMonitorDevice::Positions::print_formation(ostream & out, bool left_half, bool left_team) {
  int start_idx= 0;
  double dum= 0.0;
  if (left_half) {
    out << "\n[left half";
    dum= 52.5;
  }
  else
    out << "\n[right half";

  if (left_team) 
    out << ", left team]";
  else {
    out << ", right team]";
    start_idx= 11;
  }

  char buffer[200];
  for (int i= 1+start_idx; i< MAX_PLAYER+start_idx; i++) {
    float x= (player[i].pos.x + dum)/52.5;
    if (x<0.0) x= 0.0;
    if (x>1.0) x= 1.0;
    float y= (player[i].pos.y + 34.0)/68.0;
    if (y<0.0) y= 0.0;
    if (y>1.0) y= 1.0;
    sprintf(buffer,"\npos_%-2d = %.2f %.2f *",i-start_idx,x,y);
    out << buffer;
  }

  out << flush;
}
/*****************************************************************************/
/*****************************************************************************/
double x_SERVER_2_LP(double x) {
  return x;
}

double y_SERVER_2_LP(double y) {
  return -y;
}

double ang_SERVER_2_LP(double a) {
  a= (a*PI)/180.0;
  if (a < 0.0) return -a;
  return -a+2*PI;
}
/*****************************************************************************/
/*****************************************************************************/

SMonitorDevice::SMonitorDevice() {
  server_state.reset();
  //init_connection(); //vorlaeufig
}

bool SMonitorDevice::process_options(const ValueParser & vp) {
  vp.get("port", options.server_port);
  vp.get("host", options.server_host,STRING_MAX_SIZE);

#if USE_COACH
  vp.get("coach_port", options.coach_port);
#endif
  if ( vp.get("list_player_types",options.list_player_types) == 0 ) options.list_player_types= true;
  if ( vp.get("just_edit",options.just_edit) == 0 ) options.just_edit= true;
  //if ( vp.get("track_ball",options.track_ball,false) == 0 ) options.track_ball= true;

  vp.get("version", options.protocol_version);
  vp.get("player_radius",options.player_radius);
  vp.get("player_skin",options.player_skin);
  vp.get("kick_radius",options.kick_radius);
  vp.get("ball_radius",options.ball_radius);
  vp.get("ball_skin",options.ball_skin);   

#if 0
  int rgb[3];
  if ( 3 == vp.get("c_team_l",rgb,3) )   options.c_team_l= RGBcolor(rgb[0],rgb[1],rgb[2]);
  if ( 3 == vp.get("c_goalie_l",rgb,3) ) options.c_goalie_l= RGBcolor(rgb[0],rgb[1],rgb[2]);
  if ( 3 == vp.get("c_font_l",rgb,3) )   options.c_font_l= RGBcolor(rgb[0],rgb[1],rgb[2]);

  if ( 3 == vp.get("c_team_r",rgb,3) )   options.c_team_r= RGBcolor(rgb[0],rgb[1],rgb[2]);
  if ( 3 == vp.get("c_goalie_r",rgb,3) ) options.c_goalie_r= RGBcolor(rgb[0],rgb[1],rgb[2]);
  if ( 3 == vp.get("c_font_r",rgb,3) )   options.c_font_r= RGBcolor(rgb[0],rgb[1],rgb[2]);

  if ( 3 == vp.get("c_ball",rgb,3) )     options.c_ball= RGBcolor(rgb[0],rgb[1],rgb[2]);
  if ( 3 == vp.get("c_field",rgb,3) )    options.c_field= RGBcolor(rgb[0],rgb[1],rgb[2]);
  if ( 3 == vp.get("c_marigin",rgb,3) )  options.c_marigin= RGBcolor(rgb[0],rgb[1],rgb[2]);
  if ( 3 == vp.get("c_line",rgb,3) )     options.c_line= RGBcolor(rgb[0],rgb[1],rgb[2]);
  if ( 3 == vp.get("c_goal",rgb,3) )     options.c_goal= RGBcolor(rgb[0],rgb[1],rgb[2]);
#else
  char rgb_str[20];
  vp.get("c_team_l",rgb_str,20,'\0');   AsciiProcessor::get_col(rgb_str,options.c_team_l);
  vp.get("c_goalie_l",rgb_str,20,'\0'); AsciiProcessor::get_col(rgb_str,options.c_goalie_l);
  vp.get("c_font_l",rgb_str,20,'\0');   AsciiProcessor::get_col(rgb_str,options.c_font_l);
  
  vp.get("c_team_r",rgb_str,20,'\0');   AsciiProcessor::get_col(rgb_str,options.c_team_r);
  vp.get("c_goalie_r",rgb_str,20,'\0'); AsciiProcessor::get_col(rgb_str,options.c_goalie_r);
  vp.get("c_font_r",rgb_str,20,'\0');   AsciiProcessor::get_col(rgb_str,options.c_font_r);
  
  vp.get("c_ball",rgb_str,20,'\0');     AsciiProcessor::get_col(rgb_str,options.c_ball);
  vp.get("c_field",rgb_str,20,'\0');    AsciiProcessor::get_col(rgb_str,options.c_field);
  vp.get("c_marigin",rgb_str,20,'\0');  AsciiProcessor::get_col(rgb_str,options.c_marigin);
  vp.get("c_line",rgb_str,20,'\0');     AsciiProcessor::get_col(rgb_str,options.c_line);
  vp.get("c_goal",rgb_str,20,'\0');     AsciiProcessor::get_col(rgb_str,options.c_goal);
#endif
  if ( vp.num_of_not_accessed_entries() ) {
    ERROR_OUT << "\nSMonitorDevice: not recognized options:";
    vp.show_not_accessed_entries(ERROR_STREAM); 
    return false;
  }

  return true;
}

bool SMonitorDevice::process_options(const char * fname) {
  ValueParser vp(fname,"SMonitorDevice");
  return process_options(vp);
}

bool SMonitorDevice::process_options(int argc, char const* const* argv) {
  ValueParser vp(argc,argv,"m_");
  return process_options(vp);
}

void SMonitorDevice:: help_char_command(ostream & o) { 
  o << "
server commands:
b = drop ball at current mouse pointer position 
c = connect to server 
l = free kick for left team
r = free kick for right team
s = start a match

monitor commands:
p = print current positions on cout
P = print current formation (for left or right half
T = toggle tracking of the ball

coach commands:
t = toggle current team (starts with left team)
i = position player (of the current team) with number i at mouse pointer position
    i can take following values: 1,2,3,4,5,6,7,8,9  
                                 A or 0 which means 10
		                 B which means 11 ";
}

void SMonitorDevice::help_options(ostream & o) {
  o << "
soccer monitor options:

-m_port   [6000]";
#if USE_COACH 
  o << "
-m_coach_port    [m_port+1]";
#endif 
  o << "
-m_host   [localhost]

-m_list_player_types [false] list the available player types on the console
                             after connecting to the server.

-m_just_edit      [false] allows editing players positions without connection 
                          to server";
  //-m_track_ball     [false] allows keeping track of the ball
  o << "
-m_version        [2] version 2 is supported in sserver ver. 7.00 upwards 
                  (ver. >= 7.02 is recommended to use the drag & drop of players)

-m_player_radius  [0.3]
-m_player_skin    [1]

-m_ball_radius    [" << options.ball_radius << "]
-m_ball_skin      [1]

-m_kick_radius    [1.085]

soccer monitor color opitions:

-m_c_team_l    rrggbb  [ffff00]
-m_c_goalie_l  rrggbb  [00ff00]
-m_c_font_l    rrggbb  [ff0000]

-m_c_team_r    rrggbb  [00ffff]
-m_c_goalie_r  rrggbb  [ff99ff]
-m_c_font_r    rrggbb  [00ff00]

-m_c_ball      rrggbb  [ffffff]
-m_c_field     rrggbb  [009900]
-m_c_marigin   rrggbb  [cc3333]
-m_c_line      rrggbb  [ffffff]
-m_c_goal      rrggbb  [000000]

where rr,gg,bb in {0,33,66,99,cc,ff} for \"web sure\" colors
   or rr,gg,bb in {00,...,ff} otherwise 

comment: if options are read from a file you have to skip the preceding 
         \"-m_\"
         and put all options in a block:
         [SMonitorDevice]
         c_ball= 000000 #example of setting the ball color
";
}

bool SMonitorDevice::process_char_command(BuilderBase * build, char chr, double x, double y) {
  bool redraw= false;
  static char buf[100];
  ostrstream dum_str(buf,100);
  switch (chr) {
  case 'c':
    if ( options.just_edit ) {
      cout << "\nthis key is not set in edit mode";
      break;
    }
    server.init_serv_addr(options.server_host,options.server_port);
    send_dispinit();
    server_state.reset();
    break;
  case 'l': 
    if ( options.just_edit ) {
      cout << "\nthis key is not set in edit mode";
      break;
    }
    send_dispfoul_left(Point2d(x,y));
    break;
  case 'p': 
    cout << "\n";
    server_pos.print_inline(cout);
    cout << flush;
    break;
  case 'P': 
    cout << "\n";
    server_pos.print_formation(cout,coach_state.left_team, coach_state.left_team);
    cout << flush;
    break;
  case 'T': 
    //options.toggle_track_ball();
    break;
  case 'r': 
    if ( options.just_edit ) {
      cout << "\nthis key is not set in edit mode";
      break;
    }
    send_dispfoul_right(Point2d(x,y));
    break;
  case 's':
    if ( options.just_edit ) {
      cout << "\nthis key is not set in edit mode";
      break;
    }
    //server.init_serv_addr(options.server_host,options.server_port);
    send_dispstart();
    break;
  }

  int o_id= id_invalid;
  if ( 'b' == chr ) o_id= id_ball;
  else if ( 'A' == chr || '0' == chr) o_id= 9; 
  else if ( 'B' == chr ) o_id= 10;
  else o_id= int(chr)-49;
  if ( p_valid(o_id) && !coach_state.left_team) o_id += 11;
  

  if ( o_valid(o_id) ) {
    if ( options.just_edit) {
      set_object_pos(build,o_id,Point2d(x,y));
      return true;
    }
    else {
#if 1 // if there is no connection to the server, this is needed
      send_object_pos(o_id,Point2d(x,y));
      set_object_pos(build,o_id,Point2d(x,y)); 
      return true;
#else      
      send_object_pos(o_id,Point2d(x,y));	
      return false; //don't redraw
#endif    
    }
  }
  switch (chr) {
  case 't': coach_state.toggle_team(); break;
  case 'e': 
#if 0
    if ( options.just_edit ) {
      cout << "\nthis key is not set in edit mode";
      break;
    }
    coach.send_msg("(ear on)",9);
#else
    cout << "\nthis key is not set anymore";
#endif
    break;
    //case 'm': COACH::sock.send_msg("(change_mode play_on)",22);
  }
  return redraw;
}

bool SMonitorDevice::process_mouse_button_event(BuilderBase * build, int button, int type, unsigned long time, const Point2d & pos) {
  static int o_id= id_invalid;
  static unsigned long last_time=0;
  if (button != 2)
    return false;
  
  if ( type == MOUSE_BUTTON_PRESSED ) {
    Point2d rel;
    double min= 100.0;
    int p_min= o_id;
    for (int i=0; i<MAX_PLAYER*2; i++) {
      rel.x= pos.x - server_pos.player[i].pos.x;
      rel.y= pos.y - server_pos.player[i].pos.y;
      
      //take the greater absolute value 
      double dum=  fabs(rel.x) > fabs(rel.y) ? fabs(rel.x) : fabs(rel.y);

      if ( dum < min ) {
	min= dum;
	p_min= i;
      }
    }
    rel.x= pos.x - server_pos.ball.pos.x;
    rel.y= pos.y - server_pos.ball.pos.y;
    double dum=  fabs(rel.x) > fabs(rel.y) ? fabs(rel.x) : fabs(rel.y);
    if ( dum < min ) {
      min= dum;
      p_min= id_ball;
    }
    if ( min <= options.kick_radius * 1.5 * options.scale_factor() ) {
      o_id= p_min; 

      if ( p_valid(o_id) )
	server_pos.player[o_id].dragging= true;
      else if ( o_ball(o_id) )
	server_pos.ball.dragging= true;
      build->set_cmd_set_frame_pos(frame_shadow,pos);
      build->set_cmd_set_frame_visible(frame_shadow,1);
    }
    //cout << "\n>>>>>>>>>>>>>>>>>>>> min= " << min << " " << options.kick_radius * 1.5;
    return false; //no need for redraw
  }

  if ( type == MOUSE_BUTTON_MOTION ) 
    if ( p_valid(o_id) && abs(time-last_time)> 75 ) {
      //cout << "\n o_id= " << o_id;
      build->set_cmd_set_frame_pos(p_frame(o_id),pos);
      last_time= time;
      return true;
    }
    else if ( o_ball(o_id) && abs(time-last_time)> 75 ) {
      build->set_cmd_set_frame_pos(frame_ball,pos);
      last_time= time;
      return true;
    }
    else
      return false;
  

  if ( type == MOUSE_BUTTON_RELEASED ) {
    if ( !o_valid(o_id) ) 
      return false;
    
    bool res= false;
    if (options.just_edit) {
      set_object_pos(build,o_id,pos);
      res= true;
    }
    else {
#if 1
      set_object_pos(build,o_id,pos); //is needed 
      res= true;                       // when no connection to a server is established
#endif
      send_object_pos(o_id,pos); //send new pos to server!
    }

    if ( p_valid(o_id) )
      server_pos.player[o_id].dragging= false;
    else if ( o_ball(o_id) )
      server_pos.ball.dragging= false;

    build->set_cmd_set_frame_visible(frame_shadow,0);
    last_time= 0;
    o_id= id_invalid;
    return res;
  }

  return false;
}

bool SMonitorDevice::process_menu_button(BuilderBase * build, MenuBase * menu, int button, int mouse_button) {
  char buf[20];
  bool chg= false;
  //static unsigned long last_time=0;
  
  switch ( button ) {
  case BUTTON_START:
    send_dispstart();
    return false;

  case BUTTON_RECONNECT:
    send_dispinit();
    return false;

  case BUTTON_SCALE_LEVEL:
    if (mouse_button==1) {
      chg= false;
      if (options.scale_level > options.scale_level_min) {
	options.scale_level--;
	chg= true;
      }
    }
    else {
      if (options.scale_level < options.scale_level_max) {
	options.scale_level++;
	chg= true;
      }
    }
    if (chg) {
      double scale= options.scale_factor();
      sprintf(buf,"scale %.1f", scale);
      menu->set_button_label(button,buf);
      set_all_objects_scale(build, scale);
    }
    return chg;

  case BUTTON_INFO_LEVEL:
    if (mouse_button==1) {
      chg= false;
      if (options.info_level > options.info_level_min) {
	options.info_level--;
	chg= true;
      }
    }
    else {
      if (options.info_level < options.info_level_max) {
	options.info_level++;
	chg= true;
      }
    }
    if (chg) {
      sprintf(buf,"detail %d", options.info_level);
      menu->set_button_label(button,buf);
    }
    return chg;

  case BUTTON_QUIT:
    menu->set_exit_program();
    return false;

  }
  //cout << "\nSMonitorDev:got button " << button << flush;
  return false;
}  

bool SMonitorDevice::init_menu(MenuBase * menu) {
  char buf[20];
  menu->set_number_of_buttons(BUTTON_NUMBER);
  menu->set_button_label(BUTTON_START,"start");
  menu->set_button_label(BUTTON_RECONNECT,"(re)connect");
  double scale= options.scale_factor();
  sprintf(buf,"scale %.1f", scale);
  menu->set_button_label(BUTTON_SCALE_LEVEL,buf);
  sprintf(buf,"detail %d", options.info_level);
  menu->set_button_label(BUTTON_INFO_LEVEL,buf);
  menu->set_button_label(BUTTON_QUIT,"quit");

  //menu->set_menu_extended(false);
  menu->set_menu_extended(true);
  return true;
};

bool SMonitorDevice::init_frames(BuilderBase * build){
  const double scale= 1.0;
  const double FIELD_LENGTH                = 105.0    /scale;
  const double FIELD_WIDTH                 = 68.0    /scale;
  //const double FIELD_MARGIN                = 5.0    /scale;
  const double FIELD_CENTER_CIRCLE_R       = 9.15    /scale;
  const double FIELD_PENALTY_AREA_LENGTH   = 16.5    /scale;
  const double FIELD_PENALTY_AREA_WIDTH    = 40.32    /scale;
  const double FIELD_GOAL_AREA_LENGTH      = 5.5    /scale;
  const double FIELD_GOAL_AREA_WIDTH       = 18.32    /scale;
  const double FIELD_GOAL_WIDTH            = 14.02/scale;////7.32    /scale;
  const double FIELD_GOAL_DEPTH            = 2.44    /scale;
  const double FIELD_PENALTY_SPOT_DIST     = 11.0    /scale;
  const double FIELD_CORNER_ARC_R          = 1.0    /scale;//!!!!
  //const double FIELD_CORNER_KICK_MARGIN    = 1.0    /scale; 

  double hlen= FIELD_LENGTH * 0.5;
  double hwid= FIELD_WIDTH * 0.5;

  double pen_x= hlen- FIELD_PENALTY_AREA_LENGTH;
  double pen_y= FIELD_PENALTY_AREA_WIDTH/2.0;
  double goal_x= hlen-FIELD_GOAL_AREA_LENGTH;
  double goal_y= FIELD_GOAL_AREA_WIDTH/2.0;

  int i= 1;
  int layer= -1;

  RGBcolor c_white(255,255,255);
  RGBcolor c_black(0,0,0);

  
  Multi<Point2d> points;
  Multi<Line2d> lines;
  Multi<CircleArc2d> circlearcs;
  points.set_max_size(10);
  lines.set_max_size(25);
  circlearcs.set_max_size(10);
  
  int k= 0;
#if 0
  //field
  k= 0;
  points.tab[k++]= Point2d(-(hlen+marg),-(hwid+marg));
  points.tab[k++]= Point2d( (hlen+marg),-(hwid+marg));
  points.tab[k++]= Point2d( (hlen+marg), (hwid+marg));
  points.tab[k++]= Point2d(-(hlen+marg), (hwid+marg));
  points.set_cur_size(k);
  build->set_cmd_insert_f_polygon(0,i++,points,layer-1,options.c_field);
  //marigin

  k= 0;
  double marg= 1.2*FIELD_GOAL_WIDTH;
  marg+= 5.0;
  points.tab[k++]= Point2d(-(hlen+marg),-(hwid+marg));
  points.tab[k++]= Point2d( (hlen+marg),-(hwid+marg));
  points.tab[k++]= Point2d( (hlen+marg), (hwid+marg));
  points.tab[k++]= Point2d(-(hlen+marg), (hwid+marg));
  points.set_cur_size(k);
  build->set_cmd_insert_f_polygon(0,i++,points,layer-2,options.c_marigin);
#else
  build->set_cmd_set_background_color(options.c_field);
#endif
  //penalty points
  k= 0;
  points.tab[k++]= Point2d(-hlen+FIELD_PENALTY_SPOT_DIST,0);
  points.tab[k++]= Point2d( hlen-FIELD_PENALTY_SPOT_DIST,0);
  points.set_cur_size(k);
  build->set_cmd_insert_points(0,i++,points,layer,options.c_line);
  //lines
  k= 0;
  //field lines
  lines.tab[k++]= Line2d( Point2d(-hlen,-hwid),Point2d( hlen,-hwid));
  lines.tab[k++]= Line2d( Point2d( hlen,-hwid),Point2d( hlen,+hwid));
  lines.tab[k++]= Line2d( Point2d( hlen,+hwid),Point2d(-hlen,+hwid));
  lines.tab[k++]= Line2d( Point2d(-hlen,+hwid),Point2d(-hlen,-hwid));
  //middle line
  lines.tab[k++]= Line2d( Point2d(    0,-hwid),Point2d(    0,+hwid));
  //left penalty area
  lines.tab[k++]= Line2d( Point2d(-hlen,-pen_y),Point2d(-pen_x,-pen_y));
  lines.tab[k++]= Line2d( Point2d(-pen_x,-pen_y),Point2d(-pen_x, pen_y));
  lines.tab[k++]= Line2d( Point2d(-pen_x, pen_y),Point2d(-hlen, pen_y));
  //left goal area
  lines.tab[k++]= Line2d( Point2d(-hlen,-goal_y),Point2d(-goal_x,-goal_y));
  lines.tab[k++]= Line2d( Point2d(-goal_x,-goal_y),Point2d(-goal_x, goal_y));
  lines.tab[k++]= Line2d( Point2d(-goal_x, goal_y),Point2d(-hlen, goal_y));
  //right penalty area
  lines.tab[k++]= Line2d( Point2d( hlen,-pen_y),Point2d( pen_x,-pen_y));
  lines.tab[k++]= Line2d( Point2d( pen_x,-pen_y),Point2d( pen_x, pen_y));
  lines.tab[k++]= Line2d( Point2d( pen_x, pen_y),Point2d( hlen, pen_y));
  //right goal area
  lines.tab[k++]= Line2d( Point2d( hlen,-goal_y),Point2d( goal_x,-goal_y));
  lines.tab[k++]= Line2d( Point2d( goal_x,-goal_y),Point2d( goal_x, goal_y));
  lines.tab[k++]= Line2d( Point2d( goal_x, goal_y),Point2d( hlen, goal_y));

  lines.set_cur_size(k);
  build->set_cmd_insert_lines(0,i++,lines,layer,options.c_line);

  //cirlces
  k= 0;
  //center circle
  circlearcs.tab[k++]= CircleArc2d(Point2d(0.0,0.0),FIELD_CENTER_CIRCLE_R,0.0,2.0*M_PI);
  //edge circlearcs
  circlearcs.tab[k++]= CircleArc2d(Point2d(-hlen,-hwid),FIELD_CORNER_ARC_R,0.0,0.5*M_PI);
  circlearcs.tab[k++]= CircleArc2d(Point2d( hlen,-hwid),FIELD_CORNER_ARC_R,0.5*M_PI,M_PI);
  circlearcs.tab[k++]= CircleArc2d(Point2d( hlen, hwid),FIELD_CORNER_ARC_R,M_PI,1.5*M_PI);
  circlearcs.tab[k++]= CircleArc2d(Point2d(-hlen, hwid),FIELD_CORNER_ARC_R,1.5*M_PI,2.0*M_PI);

  circlearcs.set_cur_size(k);
  build->set_cmd_insert_circlearcs(0,i++,circlearcs,layer,options.c_line);

  //left goal
  k= 0;
  points.tab[k++]= Point2d(-hlen,-FIELD_GOAL_WIDTH*0.5);
  points.tab[k++]= Point2d(-(hlen+ FIELD_GOAL_DEPTH),-FIELD_GOAL_WIDTH*0.5);
  points.tab[k++]= Point2d(-(hlen+ FIELD_GOAL_DEPTH),+FIELD_GOAL_WIDTH*0.5);
  points.tab[k++]= Point2d(-hlen,FIELD_GOAL_WIDTH*0.5);
  points.set_cur_size(k);
  build->set_cmd_insert_f_polygon(0,i++,points,layer-1,options.c_goal);
  //right goal
  k= 0;
  points.tab[k++]= Point2d(hlen,-FIELD_GOAL_WIDTH*0.5);
  points.tab[k++]= Point2d(hlen+ FIELD_GOAL_DEPTH,-FIELD_GOAL_WIDTH*0.5);
  points.tab[k++]= Point2d(hlen+ FIELD_GOAL_DEPTH,+FIELD_GOAL_WIDTH*0.5);
  points.tab[k++]= Point2d(hlen,FIELD_GOAL_WIDTH*0.5);
  points.set_cur_size(k);
  build->set_cmd_insert_f_polygon(0,i++,points,layer-1,options.c_goal);


  double p_rad= options.player_radius/scale;
  double b_rad= options.ball_radius/scale;
  double k_rad= options.kick_radius/scale;

  //
  build->set_cmd_insert_frame(0,frame_canvas,Point2d(0.0,0.0),0.0,layer+4);
  //ball
  server_pos.set_ball_pos(Point2d(2.0,0.0));
  build->set_cmd_insert_frame(0,frame_ball,Point2d(2.0,0.0),0.0,layer+3);

  if ( options.ball_skin == 2) {
    build->set_cmd_insert_f_circle(frame_ball,0,Circle2d(Point2d(0.0,0.0),b_rad),0,options.c_ball);
    build->set_cmd_insert_circle(frame_ball,1,Circle2d(Point2d(0.0,0.0),b_rad),0,options.c_ball);
  }
  else {
    build->set_cmd_insert_f_circle(frame_ball,0,Circle2d(Point2d(0.0,0.0),b_rad),0,options.c_ball);
    build->set_cmd_insert_point(frame_ball,1,Point2d(0.0,0.0),0,options.c_ball);
    //build->set_cmd_insert_circle(frame_ball,2,Circle2d(Point2d(0.0,0.0),b_rad*10),0,options.c_ball);
    build->set_cmd_insert_circle(frame_ball,2,Circle2d(Point2d(0.0,0.0),k_rad*0.8),0,options.c_ball);
  }
  //shadow
  build->set_cmd_insert_frame(0,frame_shadow,Point2d(2.0,0.0),0.0,layer+3);
  build->set_cmd_insert_circle(frame_shadow,0,Circle2d(Point2d(0.0,0.0),k_rad),0,c_black);
  build->set_cmd_insert_line(frame_shadow,0,Line2d(Point2d(-k_rad,0.0),Point2d(k_rad,0.0)),0,c_black);
  build->set_cmd_insert_line(frame_shadow,0,Line2d(Point2d(0.0,-k_rad),Point2d(0.0,k_rad)),0,c_black);
  build->set_cmd_set_frame_visible(frame_shadow,0);
  //players
  for (int i= 0; i<MAX_PLAYER*2; i++) {
    RGBcolor c_player= options.c_invalid_l;
    RGBcolor c_player_real= options.c_team_l;
    RGBcolor c_font= options.c_font_l;
    //RGBcolor c_player_in= c_left_in;
    Point2d pos_player= Point2d(-8.0 -i*3.0*k_rad,1.1*hwid);
    if ( p_right(i) ) {
      c_player= options.c_invalid_r;
      c_player_real= options.c_team_r;
      c_font= options.c_font_r;
      //c_player_in= c_right_in;
      pos_player= Point2d(8.0 + (i-MAX_PLAYER)*3.0*k_rad,1.1*hwid);
    }
    server_pos.set_player(i,pos_player,0.0);
    build->set_cmd_insert_frame(0,p_frame(i),pos_player,0.0,layer+1);
    build->set_cmd_insert_frame(0,p_kick_frame(i),pos_player,0.0,layer+2);
    build->set_cmd_set_frame_visible(p_kick_frame(i),0);

    if (options.protocol_version >= 2) {     //insert frame for the head angle
      build->set_cmd_insert_frame(p_frame(i),p_head_frame(i),Point2d(0.0,0.0),0.0,layer+4);
      build->set_cmd_insert_line(p_head_frame(i),0,Line2d(Point2d(0.0,0.0),Point2d(k_rad,0.0)),0,RGBcolor(255,0,0));
    }

    if (options.player_skin == 2) {
      build->set_cmd_insert_f_circle(p_frame(i),1,
				  Circle2d(Point2d(0.0,0.0),p_rad),layer,c_player);
      build->set_cmd_insert_f_circlearc(p_frame(i),2,
				     CircleArc2d(Point2d(0.0,0.0),p_rad,0.5*M_PI,1.5*M_PI),
				     layer+1,c_black);
      build->set_cmd_insert_circle(p_frame(i),3,
				Circle2d(Point2d(0.0,0.0),p_rad),layer+2,c_player);//_real);

      //build->set_cmd_insert_line(p_frame(i),3,Line2d(Point2d(0.0,-p_rad),Point2d(0.0,p_rad)),layer+2,c_black);

      build->set_cmd_insert_circle(p_kick_frame(i),1,Circle2d(Point2d(0.0,0.0),p_rad),layer+2,c_white);
    }
    else {
      if (k_rad > p_rad) {
	build->set_cmd_insert_f_circle(p_frame(i),1,
				    Circle2d(Point2d(0.0,0.0),k_rad),layer,c_player);
	build->set_cmd_insert_circle(p_frame(i),5,Circle2d(Point2d(0.0,0.0),k_rad),layer+2,c_black);
      }
      build->set_cmd_insert_line(p_frame(i),3,Line2d(Point2d(0.0,0.0),Point2d(k_rad,0.0)),layer+2,c_black);

      if (options.player_radius > 0)
	build->set_cmd_insert_circle(p_frame(i),4,Circle2d(Point2d(0.0,0.0),p_rad),layer+2,c_black);

      build->set_cmd_insert_circle(p_kick_frame(i),1,Circle2d(Point2d(0.0,0.0),p_rad),layer+2,c_white);
      build->set_cmd_insert_circle(p_kick_frame(i),2,Circle2d(Point2d(0.0,0.0),k_rad),layer+2,c_white);
    }

#if 0    
    Multi<char> chr;
    chr.set_max_size(1);
    chr.set_cur_size(1);
    int k= p_number(i);
    chr.tab[0] = char (k+48);
    if (k==10) chr.tab[0] = 'A';
    if (k==11) chr.tab[0] = 'B';
    build->set_cmd_insert_string(p_frame(i),6,/* -p_rad */Point2d(0.0,0.0),chr,layer+3,c_font);
#else
    VisualString2d * v_str= new VisualString2d(0,layer+2);
    v_str->set_color(c_font);
    v_str->content.set_max_size(20);
    v_str->content.tab[0]= p_char(i);
    v_str->content.set_cur_size(1);

    server_pos.vis_player[i].str= v_str;
    build->set_cmd_insert_visobject(p_frame(i),server_pos.vis_player[i].str);
#endif
  }
  return true;
}

bool SMonitorDevice::init_connection() {
  if (options.just_edit)
    return true;

  server.init_socket_fd();
  server.set_fd_nonblock();
  server.init_serv_addr(options.server_host,options.server_port);

#if USE_COACH
  coach.init_socket_fd();
  coach.set_fd_nonblock();
  coach.init_serv_addr(options.server_host,options.coach_port);
#endif  

  send_dispinit();
  return true;
}

bool SMonitorDevice::process_input(fd_set * set, BuilderBase * build) {
  if (options.just_edit) 
    return true;

  char * buf;

  bool redraw= false;
  //process server commands
  if ( is_fd_in_set(server.socket_fd, set) ) {
    int num_bytes= 0;
    int msg_type;
    int num_show_msg= 0;

    buf= buffer1;
    while ( server.recv_msg( buf , num_bytes, true ) ) {
      //if (RUN::freeze) continue;
      msg_type= server_msg_type(buf);
      if (SHOW_MODE == msg_type) {
	if (num_bytes > 0) {
	  num_show_msg++;
	  //rotate the buf to keep the last show message in the other buffer!
	  if (buf == buffer1)
	    buf= buffer2;
	  else 
	    buf= buffer1;
	}
      }
      else if (MSG_MODE == msg_type) {
	bool res= server_interpret_msginfo_t(build, (void*)(buf));
	redraw= redraw || res;
      }
      else if (DRAW_MODE == msg_type) {
	bool res= server_interpret_drawinfo_t(build, (void*)(buf));
	redraw= redraw || res;
      }
      else if (NO_INFO == msg_type)
	cerr << "\nmode= NO_INFO not yet implemented";
      else if (BLANK_MODE == msg_type)
	cerr << "\nmode= BLANK_MODE not yet implemented";
      else if (PM_MODE == msg_type)
	cerr << "\nmode= PM_MODE not yet implemented";
      else if (TEAM_MODE == msg_type)
	cerr << "\nmode= TEAM_MODE not yet implemented";
      else if (PT_MODE == msg_type)
	return server_interpret_player_type_t(build, (void*)(buf));
      else if (PARAM_MODE == msg_type)
	cerr << "\nmode= PARAM_MODE not yet implemented";
      else if (PPARAM_MODE == msg_type)
	cerr << "\nmode= PPARAM_MODE not yet implemented";
      else {
	ERROR_OUT << "\nunknown mode " << msg_type << " :";
	int i=0;
	while(i<50) {
	  if (isprint(buf[i])) 
	    cout << buf[i];
	  else 
	    break;
	  i++;
	}
	if (i>0) {
	  cout << "\nHINT: if you use server version < 7.02 repositioning of players"
	       << "\nmay couse such messages" << flush;
	}
      }
    }
    if ( num_show_msg > 0) { 
      //switch to the cached buffer
      if (buf == buffer1)
	buf= buffer2;
      else 
	buf= buffer1;

      bool res;
      if (options.protocol_version >= 2) 
	res= server_interpret_showinfo_t2(build, (void*)(buf));
      else
	res= server_interpret_showinfo_t(build, (void*)(buf));
      redraw= redraw || res;
      if (num_show_msg>1)
	cerr << "\nskipped " << num_show_msg-1 << " SHOW messages";
    }
  }
   
#if USE_COACH
  //process coach commands
  if ( is_fd_in_set( coach.socket_fd, set ) ) {
    int num_bytes;
    while ( coach.recv_msg( buffer[0] , num_bytes, true ) ) {
      buffer[0][num_bytes]= '\0';
      cerr << "\ncoach received: " << buffer[0] << endl;
    }
  }
#endif

  return redraw;
}

int  SMonitorDevice::set_fds(fd_set * set) { 
  if (options.just_edit)
    return -1;
  add_fd_to_set(server.socket_fd,set); 
#if USE_COACH
  add_fd_to_set(coach.socket_fd,set); 
  return server.socket_fd>coach.socket_fd?server.socket_fd:coach.socket_fd; 
#else
  return server.socket_fd; 
#endif

}

bool SMonitorDevice::got_fds(fd_set * set) { 
  if (options.just_edit)
    return false;
  if ( is_fd_in_set(server.socket_fd,set) ) return true;
#if USE_COACH
  return is_fd_in_set(coach.socket_fd,set);
#else
  return false;
#endif
}


/* private methods */

int SMonitorDevice::server_msg_type(void * ptr) {
  SSrv::dispinfo_t *dispinfo_dum= (SSrv::dispinfo_t*)ptr;
  SSrv::dispinfo_t & dispinfo= *dispinfo_dum;
  return  (ntohs (dispinfo.mode));
}

const char * SMonitorDevice::get_2D_object_par(const char * msg,int par_num,
					       double & p1, double & p2, double & p3, double & p4, 
					       char * color_str,int color_str_max_len ) {
  //while (msg[0] == ' ') msg++;
  if (par_num > 0 ) {
    char* next;
    p1= strtod(msg,&next);
    if (next == 0) return 0;
    msg= next;
    //  cout<< "\n1:" << msg;
    if (par_num > 1) {
      p2= strtod(msg,&next);
      if (next == 0) return 0;
      msg= next;
      if ( par_num > 2) {
	p3=  strtod(msg,&next);
	if (next == 0) return 0;
	msg= next;
	if ( par_num > 3 ) {
	  p4= strtod(msg,&next);
	  if (next == 0) return 0;
	  msg= next;
	}
      }
    }
  }
  //  cout<< "\n1:" << msg;
  while (msg[0] == ' ') msg++;
  //  cout<< "\n1:" << msg;
  strncpy(color_str,msg,color_str_max_len);
  color_str[color_str_max_len-1] = ' ';
  int i= 0;
  while ( color_str[i] != ' ' && color_str[i] != '\0' && color_str[i] != '\n' ) i++;
  color_str[i] = '\0';
  msg += i;
  return msg;
}

bool SMonitorDevice::server_interpret_showinfo_t(BuilderBase * build, void *ptr) {
  static char *PlayModeString[] = PLAYMODE_STRINGS;
  SSrv::dispinfo_t *dispinfo= (SSrv::dispinfo_t*)ptr;

  if (ntohs (dispinfo->mode) != SHOW_MODE) {
    ERROR_OUT << "\nmode != SHOW_MODE";
    return false;
  }

  const SSrv::showinfo_t &showinfo= dispinfo->body.show; //shortcut
  server_state.left_teamname.set( showinfo.team[0].name);
  server_state.right_teamname.set( showinfo.team[1].name);
  //match info
  short s_l= ntohs(showinfo.team[0].score);
  short s_r= ntohs(showinfo.team[1].score);
  if ( '\0'==server_state.left_teamname.name[0] ) s_l = 0;
  if ( '\0'==server_state.right_teamname.name[0] ) s_r = 0;
	       
  char dum[30];
  char * pmstr;
  if (showinfo.pmode >= 0 && showinfo.pmode < 32)
    pmstr= PlayModeString[showinfo.pmode];
  else {
    sprintf(dum,"sim_mode %d",showinfo.pmode);
    pmstr= dum;
  }
      
  sprintf(server_state.match_information.name," %10s %d:%d %-10s %16s %6d    ",
	  showinfo.team[0].name, int(s_l),int(s_r),
	  showinfo.team[1].name,
	  pmstr,
	  ntohs(showinfo.time));
    
  if ( server_state.reconnected ) //clear the drawing plane
    build->set_cmd_empty_frame(frame_canvas);
	
  { //set the ball
    /* Zwischenspeicherung als short notwendig, da ansonsten ntohs nicht richtig funkioniert */
    short s_x      = ntohs(showinfo.pos[0].x);
    short s_y      = ntohs(showinfo.pos[0].y);
    Point2d pos;
    pos.x     = x_SERVER_2_LP (s_x / SHOWINFO_SCALE);
    pos.y     = y_SERVER_2_LP (s_y / SHOWINFO_SCALE);
    server_pos.set_ball_pos(pos); //just to collect informations
    if ( !server_pos.ball.dragging )
      build->set_cmd_set_frame_pos(frame_ball,pos);
    else
      build->set_cmd_set_frame_pos_ang(frame_shadow,pos,Angle());

    if (options.track_ball)
      build->set_cmd_set_frame_pos(0,Point2d(- pos.x,-pos.y)); 
  } 

  for (int i= 0; i< 2*MAX_PLAYER; i++) { //players [0 ... MAX_PLAYER*2-1] 
    /* Zwischenspeicherung als short notwendig, da ansonsten ntohs nicht richtig funkioniert */
    short s_x      = ntohs(showinfo.pos[i+1].x);
    short s_y      = ntohs(showinfo.pos[i+1].y);
    short s_enable = ntohs(showinfo.pos[i+1].enable);
    //short s_side   = ntohs(showinfo.pos[i+1].side);
    //short s_unum   = ntohs(showinfo.pos[i+1].unum);
    short s_angle  = ntohs(showinfo.pos[i+1].angle);
    Point2d pos;
    pos.x     = x_SERVER_2_LP (s_x / SHOWINFO_SCALE);
    pos.y     = y_SERVER_2_LP (s_y / SHOWINFO_SCALE);
    Angle angle = Angle( ang_SERVER_2_LP ( double(s_angle)) );

    server_pos.set_player(i,pos,angle); //just to collect informations
    if ( !server_pos.player[i].dragging )
      build->set_cmd_set_frame_pos_ang(p_frame(i),pos,angle);
    else
      build->set_cmd_set_frame_pos_ang(frame_shadow,pos,angle);

    if ( s_enable == DISABLE ) {
      if ( server_pos.player_alive(i) || server_state.reconnected ) { //disable the player just once!
	//cout << "\nchanging color disabled " << i;
	server_pos.set_player_alive(i,false);
	if ( p_left(i) )  //left team
	  build->set_cmd_set_object_color(p_frame(i),1,options.c_invalid_l);
	else
	  build->set_cmd_set_object_color(p_frame(i),1,options.c_invalid_r);
      }
      continue;
    }
    //here the player is enabled!!!
    if (!server_pos.player_alive(i) || server_state.reconnected ) {
      server_pos.set_player_alive(i);
      // initialize colors
      if ( p_left(i) ) { //left team
	if ( s_enable & GOALIE )
	  build->set_cmd_set_object_color(p_frame(i),1,options.c_goalie_l);
	else
	  build->set_cmd_set_object_color(p_frame(i),1,options.c_team_l);
      }
      else //right team
	if ( s_enable & GOALIE )
	  build->set_cmd_set_object_color(p_frame(i),1,options.c_goalie_r);
	else
	  build->set_cmd_set_object_color(p_frame(i),1,options.c_team_r);
    }

    if ( s_enable & KICK ) {
      //cout << "  <------------ KICK";
      build->set_cmd_set_frame_visible(p_kick_frame(i),1);
      build->set_cmd_set_frame_pos(p_kick_frame(i),pos);
    }
    else
      build->set_cmd_set_frame_visible(p_kick_frame(i),0);
  }
  server_state.reconnected= false;
  return true;
}


bool SMonitorDevice::server_interpret_drawinfo_t(BuilderBase * build, void *ptr) {
  SSrv::dispinfo_t *dispinfo= (SSrv::dispinfo_t*)ptr;
  
  short s_x, s_y, s_x2, s_y2, s_r;      
  Point2d pos, pos2;
  double r;
  bool redraw= false;
  const int idx_zero = 0;

  if (ntohs (dispinfo->mode) != DRAW_MODE) {
    ERROR_OUT << "\nmode != DRAW_MODE";
    return false;
  }

  const SSrv::drawinfo_t &drawinfo= dispinfo->body.draw; //shortcut
  switch (ntohs(drawinfo.mode)) {
  case DrawPoint:
    redraw= true;
    s_x   = ntohs(drawinfo.object.pinfo.x);
    s_y   = ntohs(drawinfo.object.pinfo.y);
    pos.x     = x_SERVER_2_LP (s_x );
    pos.y     = y_SERVER_2_LP (s_y );
    //cout << "\nDrawPoint " << draw_obj_idx << " x,y= " << x << "," << y;
    build->set_cmd_insert_point(frame_canvas,idx_zero,pos,0,
				RGB_DB::XNamedColor_to_RGBcolor(drawinfo.object.pinfo.color));

    break ;
  case DrawCircle:
    redraw= true;
    s_x   = ntohs(drawinfo.object.cinfo.x);
    s_y   = ntohs(drawinfo.object.cinfo.y);
    s_r   = ntohs(drawinfo.object.cinfo.r);
    r     = s_r;
    pos.x     = x_SERVER_2_LP ( s_x ) + r;
    pos.y     = y_SERVER_2_LP ( s_y ) - r;

    //cout << "\nDrawCircle " << draw_obj_idx << " (x,y),r= (" << x << "," << y << "), "<< r;
    build->set_cmd_insert_circle(frame_canvas,idx_zero,Circle2d(pos,r),0,
				 RGB_DB::XNamedColor_to_RGBcolor(drawinfo.object.cinfo.color));
    break ;
  case DrawLine:
    redraw= true;
    s_x   = ntohs(drawinfo.object.linfo.x1);
    s_y   = ntohs(drawinfo.object.linfo.y1);
    s_x2  = ntohs(drawinfo.object.linfo.x2);
    s_y2  = ntohs(drawinfo.object.linfo.y2);
    pos.x     = x_SERVER_2_LP (s_x  );
    pos.y     = y_SERVER_2_LP (s_y  );
    pos2.x    = x_SERVER_2_LP (s_x2 );
    pos2.y    = y_SERVER_2_LP (s_y2 );
    build->set_cmd_insert_line(frame_canvas,idx_zero,Line2d(pos,pos2),0,
			       RGB_DB::XNamedColor_to_RGBcolor(drawinfo.object.linfo.color));
    break ;
  case DrawClear:
    redraw= true;
    build->set_cmd_empty_frame(frame_canvas);
    //field.drawclear() ;
    break ;
  default:
    ERROR_OUT << "\nIllegal value in DRAW_MODE: " << ntohs(drawinfo.mode);
  }
  return redraw;
}

//>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
bool SMonitorDevice::server_interpret_msginfo_t(BuilderBase * build, void *ptr) {
  SSrv::dispinfo_t *dispinfo= (SSrv::dispinfo_t*)ptr;
  const int idx_zero= 0;

  if (ntohs (dispinfo->mode) != MSG_MODE) {
    ERROR_OUT << "\nmode != MSG_MODE";
    return false;
  }
  //cout <<"\nMSG_MODE ";
  double KONV= 1.0;
  //left player is sending on MSG_BOARD, right player is sending on LOG_BOARD
  //coordinates of right player have to be converted
  if (dispinfo->body.msg.board == LOG_BOARD)
    KONV= -1.0;

  const char* msg= dispinfo->body.msg.message; //shortcut
  if ( strncmp(msg,"_2D_",4) != 0) return false;
  msg += 4;

  char color[100];
  RGBcolor col;
  double x1,y1,x2,y2;

  while ( msg[0] != '\0' /* && msg[0] != '\n' */) {
    while (msg[0] == ' ' || msg[0] == '\n') msg++;
    if (msg[0]=='\0')
      break;
    if (strncmp (msg,"clear ",6) == 0 ) {
      //clear drawing plane
      msg += 6;
      build->set_cmd_empty_frame(frame_canvas);
    }
    else
      if (msg[0]== 'l') {
	msg++;
	msg= get_2D_object_par(msg,4,x1,y1,x2,y2,color,100);
	//cout << "\n line " << x1 << " " << y1 << " " << x2 << " " << y2;
	if ( 0 == msg ) return false;
	if ( color[0] == '#')
	  AsciiProcessor::get_col(color+1,col);
	else
	  col = RGB_DB::XNamedColor_to_RGBcolor(color);
	build->set_cmd_insert_line(frame_canvas,idx_zero,
				   Line2d(Point2d(KONV*x1,KONV*y1),Point2d(KONV*x2,KONV*y2)),0,
				   col);
      }
      else
	if (msg[0]== 'c') {
	  msg++;
	  msg= get_2D_object_par(msg,3,x1,y1,x2,y2,color,100);
	  if ( 0 == msg ) return false;
	  if ( color[0] == '#')
	    AsciiProcessor::get_col(color+1,col);
	  else
	    col = RGB_DB::XNamedColor_to_RGBcolor(color);
	  build->set_cmd_insert_circle(frame_canvas,idx_zero,
				       Circle2d(Point2d(KONV*x1,KONV*y1),x2),0,
				       col);
	}
	else
	  if (msg[0]== 'p') {
	    msg++;
	    msg= get_2D_object_par(msg,2,x1,y1,x2,y2,color,100);
	    if ( 0 == msg ) return false;
	    if ( color[0] == '#')
	      AsciiProcessor::get_col(color+1,col);
	    else
	      col = RGB_DB::XNamedColor_to_RGBcolor(color);
	    build->set_cmd_insert_point(frame_canvas,idx_zero,
					Point2d(KONV*x1,KONV*y1),0,
					col);
	  }
	  else {
	    return false;
	  }
  }
  return true;
  //cout << dispinfo.body.msg.message << flush;
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
bool SMonitorDevice::server_interpret_showinfo_t2(BuilderBase * build, void *ptr) {
  static char *PlayModeString[] = PLAYMODE_STRINGS;
  SSrv::dispinfo_t2 *dispinfo= (SSrv::dispinfo_t2*)ptr;

  if (ntohs (dispinfo->mode) != SHOW_MODE) {
    ERROR_OUT << "\nmode != SHOW_MODE";
    return false;
  }

  const SSrv::showinfo_t2 &showinfo= dispinfo->body.show; //shortcut
  server_state.left_teamname.set( showinfo.team[0].name);
  server_state.right_teamname.set( showinfo.team[1].name);
  //match info
  short s_l= ntohs(showinfo.team[0].score);
  short s_r= ntohs(showinfo.team[1].score);
  if ( '\0'==server_state.left_teamname.name[0] ) s_l = 0;
  if ( '\0'==server_state.right_teamname.name[0] ) s_r = 0;
	       
  char dum[30];
  char * pmstr;
  if (showinfo.pmode >= 0 && showinfo.pmode < 32)
    pmstr= PlayModeString[showinfo.pmode];
  else {
    sprintf(dum,"sim_mode %d",showinfo.pmode);
    pmstr= dum;
  }
      
  sprintf(server_state.match_information.name," %10s %d:%d %-10s %16s %6d    ",
	  showinfo.team[0].name, int(s_l),int(s_r),
	  showinfo.team[1].name,
	  pmstr,
	  ntohs(showinfo.time));
    
  if ( server_state.reconnected ) //clear the drawing plane
    build->set_cmd_empty_frame(frame_canvas);

  //set players
  for (int i= 0; i< 2*MAX_PLAYER; i++) {
    /* Zwischenspeicherung als short notwendig, da ansonsten ntohs nicht richtig funkioniert */
    long l_x      = ntohl(showinfo.pos[i].x);
    long l_y      = ntohl(showinfo.pos[i].y);
    short s_mode  = ntohs(showinfo.pos[i].mode);
    //short s_type  = ntohs(showinfo.pos[i].type);
    long l_body_angle  = ntohl(showinfo.pos[i].body_angle);
    long l_head_angle  = ntohl(showinfo.pos[i].head_angle);
    Point2d pos;
    pos.x     = x_SERVER_2_LP ( double(l_x) / SHOWINFO_SCALE2);
    pos.y     = y_SERVER_2_LP ( double(l_y) / SHOWINFO_SCALE2);

    //Angle body_angle = ang_SERVER_2_LP ( double(l_body_angle) / SHOWINFO_SCALE2);
    //angles are now radians (but in wrong direction) (ver. 7.00),
    Angle body_angle =  - double(l_body_angle) / SHOWINFO_SCALE2;
    Angle head_angle =  - double(l_head_angle) / SHOWINFO_SCALE2;
    /*
     if (false && i == 0)
      cout << "\nl_body_angle= " << l_body_angle
	   << ", res= " << double(l_body_angle) / SHOWINFO_SCALE2
	   << ", res2= " << ang_SERVER_2_LP ( double(l_body_angle) / SHOWINFO_SCALE2) 
	   << "\nl_head_angle= " << l_head_angle
	   << ", res= " << double(l_head_angle) / SHOWINFO_SCALE2
	   << ", res2= " << ang_SERVER_2_LP ( double(l_head_angle) / SHOWINFO_SCALE2) << flush;
   
    if (false && i == 0) {
      cout << "\n-- time (" << ntohs(showinfo.time) << ")"
	   << "\ntype= " << ntohs(showinfo.pos[i].type)
	   << "\nstamina= " << double(ntohl(showinfo.pos[i].stamina)) /SHOWINFO_SCALE2
	   << "\neffort= " << double(ntohl(showinfo.pos[i].effort)) / SHOWINFO_SCALE2
	   << "\nrecovery= " << double(ntohl(showinfo.pos[i].recovery)) / SHOWINFO_SCALE2
	   << "\nk= " << ntohs(showinfo.pos[i].kick_count) 
	   << " , d= " << ntohs(showinfo.pos[i].dash_count) 
	   << " , t= " << ntohs(showinfo.pos[i].turn_count) 
	   << " , m= " << ntohs(showinfo.pos[i].move_count) 
	   << " ; c= " << ntohs(showinfo.pos[i].catch_count) 
	   << " | n= " << ntohs(showinfo.pos[i].tneck_count) 
	   << " | s= " << ntohs(showinfo.pos[i].say_count) 
	   << " | v= " << ntohs(showinfo.pos[i].chg_view_count);
    }
    */
    //server_pos.player[i].stamina= double(ntohl(showinfo.pos[i].stamina)) /SHOWINFO_SCALE2;

    server_pos.set_player(i,pos,Angle(body_angle)); //just to collect informations
    if ( options.info_stamina() ) {
      char * buf= server_pos.vis_player[i].str->content.tab;
      long ddd= ntohl(showinfo.pos[i].stamina) / SHOWINFO_SCALE2_int;
      sprintf(buf+1," %d", (int)ddd);
      server_pos.vis_player[i].str->content.set_cur_size(strlen(buf));
    }
    else // not yet really ok
      server_pos.vis_player[i].str->content.set_cur_size(1);

    if ( !server_pos.player[i].dragging ) {
      build->set_cmd_set_frame_pos_ang(p_frame(i),pos,body_angle);
      build->set_cmd_set_frame_ang(p_head_frame(i),head_angle);
    }
    else
      build->set_cmd_set_frame_pos_ang(frame_shadow,pos,body_angle);

    if ( s_mode == DISABLE ) {
      if ( server_pos.player_alive(i) || server_state.reconnected ) { //disable the player just once!
	//cout << "\nchanging color disabled " << idx;
	server_pos.set_player_alive(i,false);
	if ( p_left(i) )  //left team
	  build->set_cmd_set_object_color(p_frame(i),1,options.c_invalid_l);
	else
	  build->set_cmd_set_object_color(p_frame(i),1,options.c_invalid_r);
      }
      continue;
    }
    //here the player is enabled!!!
    if (!server_pos.player_alive(i) || server_state.reconnected ) {
      server_pos.set_player_alive(i);
      //cout << "\n set player alive " << idx;
      // initialize colors
      if ( p_left(i) ) { //left team
	if ( s_mode & GOALIE )
	  build->set_cmd_set_object_color(p_frame(i),1,options.c_goalie_l);
	else
	  build->set_cmd_set_object_color(p_frame(i),1,options.c_team_l);
      }
      else //right team
	if ( s_mode & GOALIE )
	  build->set_cmd_set_object_color(p_frame(i),1,options.c_goalie_r);
	else
	  build->set_cmd_set_object_color(p_frame(i),1,options.c_team_r);
    }
    
    if ( s_mode & KICK ) {
      //cout << "  <------------ KICK";
      build->set_cmd_set_frame_visible(p_kick_frame(i),1);
      build->set_cmd_set_frame_pos(p_kick_frame(i),pos);
    }
    else
      build->set_cmd_set_frame_visible(p_kick_frame(i),0);
  }
  
  //set the ball
  long l_x      = ntohl(showinfo.ball.x);
  long l_y      = ntohl(showinfo.ball.y);
  //the velocity is 
  long l_vx     = ntohl(showinfo.ball.deltax);
  long l_vy     = ntohl(showinfo.ball.deltay);
  Point2d pos,vel;
  pos.x= x_SERVER_2_LP ( double(l_x) / SHOWINFO_SCALE2);
  pos.y= y_SERVER_2_LP ( double(l_y) / SHOWINFO_SCALE2);
  vel.x= x_SERVER_2_LP ( double(l_vx) / SHOWINFO_SCALE2);
  vel.y= y_SERVER_2_LP ( double(l_vy) / SHOWINFO_SCALE2);

  server_pos.set_ball(pos,vel); //just to collect informations
  if ( !server_pos.ball.dragging )
    build->set_cmd_set_frame_pos(frame_ball,pos);
  else
    build->set_cmd_set_frame_pos_ang(frame_shadow,pos,Angle());

  if (options.track_ball)
    build->set_cmd_set_frame_pos(0,Point2d(-pos.x,-pos.y)); 
  
  server_state.reconnected= false;

  return true;
}

bool SMonitorDevice::server_interpret_player_type_t(BuilderBase * build, void *ptr) {
  SSrv::dispinfo_t2 *dispinfo= (SSrv::dispinfo_t2*)ptr;

  if (ntohs (dispinfo->mode) != PT_MODE) {
    ERROR_OUT << "\nmode != PT_MODE";
    return false;
  }
  static bool was_here= false;
  if (!was_here)
    cerr << "\nmode= PT_MODE not yet fully implemented (coming soon)";
  
  was_here= true;

  if (options.list_player_types) {
    const SSrv::player_type_t &player_type= dispinfo->body.ptinfo; //shortcut
    cout << ""
	 << "\ntype= " << ntohs(player_type.id)
	 << "\nplayer_speed_max= " << double(ntohl(player_type.player_speed_max)) / SHOWINFO_SCALE2
	 << "\nstamina_inc_max = " << double(ntohl(player_type.stamina_inc_max)) / SHOWINFO_SCALE2
	 << "\nplayer_decay    = " << double(ntohl(player_type.player_decay)) / SHOWINFO_SCALE2
	 << "\ninertia_moment  = " << double(ntohl(player_type.inertia_moment)) / SHOWINFO_SCALE2
	 << "\ndash_power_rate = " << double(ntohl(player_type.dash_power_rate)) / SHOWINFO_SCALE2
	 << "\nplayer_size     = " << double(ntohl(player_type.player_size)) / SHOWINFO_SCALE2
	 << "\nkickable_margin = " << double(ntohl(player_type.kickable_margin)) / SHOWINFO_SCALE2
	 << "\nkick_rand       = " << double(ntohl(player_type.kick_rand)) / SHOWINFO_SCALE2
      // << "\nextra_stamina   = " << double(ntohl(player_type.extra_stamina)) / SHOWINFO_SCALE2
      // << "\neffort_max      = " << double(ntohl(player_type.effort_max)) / SHOWINFO_SCALE2
      // << "\neffort_min      = " << double(ntohl(player_type.effort_min)) / SHOWINFO_SCALE2
	 << "";
  }
  return false;
}

char SMonitorDevice::p_char(int i) {
  int k= p_number(i);
  char dum;
  dum = char (k+48);
  if (k==10) dum = 'A';
  if (k==11) dum = 'B';
  return dum;
}

void SMonitorDevice::set_object_pos(BuilderBase * build, int o_id, const Point2d & pos) {
  if ( o_ball(o_id) ) {
    server_pos.set_ball_pos(pos); //just to collect informations	
    build->set_cmd_set_frame_pos(frame_ball,pos);
    return;
  }

  server_pos.set_player_pos(o_id,pos); //just to collect informations	
  build->set_cmd_set_frame_pos(p_frame(o_id),pos);

  if (!server_pos.player_alive(o_id)) {
    server_pos.set_player_alive(o_id);
    if ( p_left(o_id) ) { //left team
      if ( p_number(o_id) == 1 ) //assume goalie has number 1 in his team 
	build->set_cmd_set_object_color(p_frame(o_id),1,options.c_goalie_l);
      else 
	build->set_cmd_set_object_color(p_frame(o_id),1,options.c_team_l);
    }
    else //right team
      if ( p_number(o_id) == 1 ) //assume goalie has number 1 in his team 
	build->set_cmd_set_object_color(p_frame(o_id),1,options.c_goalie_r);
      else
	build->set_cmd_set_object_color(p_frame(o_id),1,options.c_team_r);
  }
}

void SMonitorDevice::set_all_objects_scale(BuilderBase * build, double scale) {
  build->set_cmd_set_frame_scale(frame_ball, scale);
  build->set_cmd_set_frame_scale(frame_shadow, scale);
  for (int i=0; i< 2*MAX_PLAYER; i++) {
    build->set_cmd_set_frame_scale(p_frame(i),scale);
    build->set_cmd_set_frame_scale(p_kick_frame(i),scale);
  }
}

void SMonitorDevice::send_object_pos(int o_id, const Point2d & pos) {
  static char dum_buf[100];
  if (o_ball(o_id)) {
    send_dispball(pos);
    return;
  }

  strstream dum_str(dum_buf,100);
#if 0 // #if USE_COACH for server version < 7.02
  dum_str << "(move (player ";
  if ( p_left(o_id) )
    dum_str << server_state.left_teamname.name;
  else 
    dum_str << server_state.right_teamname.name;
  dum_str << " " << p_number(o_id) << ") " << int(pos.x) << " " <<  int(-pos.y) << ")";
  coach.send_msg(dum_str.str(),dum_str.pcount());
#else 
  /* don't use the coach to position players but the
     (dispplayer ...) command (works in server 7.02 upwards) */
  dum_str << "(dispplayer";
  if ( p_left(o_id) )
    dum_str << " 1";
  else 
    dum_str << " -1";

  //preserve the angle of the player
  int tmp= int(-(server_pos.player[o_id].ang.get_value()- 2*M_PI) * 180.0/M_PI);
  if (tmp > 180)
    tmp-= 360;
  //cout << "\n body agnle = " << int (tmp) << flush; //don't change angle 

  dum_str << " " << p_number(o_id) << " " 
	  << int(pos.x * SHOWINFO_SCALE) << " " <<  int(-pos.y * SHOWINFO_SCALE) 
	  << " 0 0 "
          << tmp << ")";
  server.send_msg(dum_str.str(),dum_str.pcount());
#endif
}

void SMonitorDevice::send_dispinit() {
  if (options.protocol_version >= 2) {
    const char msg[]= "(dispinit version 2)";
    server.send_msg(msg,strlen(msg)+1);
  }
  else {
    const char msg[]= "(dispinit)";
    server.send_msg(msg,strlen(msg)+1);
    //cout << "\ntime=" << TOOLS::get_current_ms_time() << ", sending msg=:" << msg << flush;
  }
}

void SMonitorDevice::send_dispstart() {
  const char msg[]= "(dispstart)";
  server.send_msg(msg,strlen(msg)+1);
}

void SMonitorDevice::send_dispball(const Point2d & pos) {
  static char buf[100];
  sprintf(buf,"(dispfoul %d %d 0)",int(pos.x * SHOWINFO_SCALE),int( - pos.y * SHOWINFO_SCALE));
  server.send_msg(buf,strlen(buf)+1);
}

void SMonitorDevice::send_dispfoul_left(const Point2d & pos) {
  static char buf[100];
  sprintf(buf,"(dispfoul %d %d 1)",int(pos.x * SHOWINFO_SCALE),int( - pos.y * SHOWINFO_SCALE));
  server.send_msg(buf,strlen(buf)+1);
}

void SMonitorDevice::send_dispfoul_right(const Point2d & pos) {
  static char buf[100];
  sprintf(buf,"(dispfoul %d %d -1)",int(pos.x * SHOWINFO_SCALE),int( - pos.y * SHOWINFO_SCALE));
  server.send_msg(buf,strlen(buf)+1);
}

