/*
 * Copyright (c) 1999 - 2001, Artur Merke <amerke@ira.uka.de> 
 *
 * This code is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * It is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */
#include "strconv.h"
#include <stdlib.h>
#include <errno.h>
#include <limits.h>
#include <string.h>

#if 0
#define IS_WHITE_CHAR(x) ( ' ' == (x) || '\t' == (x) )
#define IS_STR_END_CHAR(x) ( '\0' == (x) || '\n' == (x) )
#else
#define IS_WHITE_CHAR(x) ( ' ' == (x) || '\t' == (x) || '\n'== (x) )
#define IS_STR_END_CHAR(x) ( '\0' == (x) )
#endif

#define IS_EXPR_END_CHAR(x) ( '\0' == (x) )

int strskip(char const* str,char const* expr, char const* & next) {
  if (0==str || 0==expr) {
    next= 0;
    return 0;
  }
    
  while ( IS_WHITE_CHAR(*str) )
    str++;

  while ( *str == *expr &&  !IS_EXPR_END_CHAR(*expr) ) {
    str++;
    expr++;
  }
  if ( !IS_EXPR_END_CHAR(*expr) || 
       !IS_STR_END_CHAR(*str) && !IS_WHITE_CHAR(*str) ) {
    next= 0;
    return 0;
  }

  while ( IS_WHITE_CHAR(*str) )
    str++;

  if ( IS_STR_END_CHAR(*str) )
    next= 0;
  else
    next= str;

  return 1;
}

int strshift(char const* str,char const* expr, char const* & next) {
  if (0==str || 0==expr) {
    next= 0;
    return 0;
  }

  while( IS_WHITE_CHAR(*str) )
    str++;
  
  while ( *str == *expr && !IS_EXPR_END_CHAR(*expr) ) {
    str++;
    expr++;
  }
  
  if ( !IS_EXPR_END_CHAR(*expr) ) {
    next= 0;
    return 0;
  }

  while ( IS_WHITE_CHAR(*str) )
    str++;

  if ( IS_STR_END_CHAR(*str) )
    next= 0;
  else
    next= str;

  return 1;
}


int _strconv(char const* str,int & val, char const* & next, bool must_have_end) {
  char* error_ptr;
  long tmp_long;

  if (0 == str) {
    next= 0;
    return 0;
  }

  tmp_long= strtol(str, &error_ptr, 10);
  next= error_ptr;

  if ( str == next 
       || ERANGE == errno 
       || tmp_long > INT_MAX || tmp_long < INT_MIN ) {
    next= 0;
    return 0;
  }


  if ( !must_have_end || IS_STR_END_CHAR(*next) || IS_WHITE_CHAR(*next) ) {
    val= tmp_long;
  }
    
  while ( IS_WHITE_CHAR(*next) )
    next++;
   
  if ( IS_STR_END_CHAR(*next) )
    next= 0;

  
  return 1;
}

int _strconv(char const* str,long & val, char const* & next, bool must_have_end) {
  char* error_ptr;
  long tmp_long;

  if (0 == str) {
    next= 0;
    return 0;
  }

  tmp_long= strtol(str, &error_ptr, 10);
  next= error_ptr;

  if ( str == next 
       || ERANGE == errno 
       || tmp_long > LONG_MAX || tmp_long < LONG_MIN ) {
    next= 0;
    return 0;
  }

  if ( !must_have_end || IS_STR_END_CHAR(*next) || IS_WHITE_CHAR(*next) ) {
    val= tmp_long;
  }

  while ( IS_WHITE_CHAR(*next) )
    next++;

  if ( IS_STR_END_CHAR(*next) )
    next= 0;
  
  return 1;
}


int _strconv(char const* str,double & val, char const* & next, bool must_have_end) {
  char* error_ptr;
  double tmp_double;

  if (0 == str) {
    next= 0;
    return 0;
  }


  tmp_double= strtod(str, &error_ptr);
  next= error_ptr;

  if ( str == next 
       || ERANGE == errno ) {
    next= 0;
    return 0;
  }

  if ( !must_have_end || IS_STR_END_CHAR(*next) || IS_WHITE_CHAR(*next) ) {
    val= tmp_double;
  }

  while ( IS_WHITE_CHAR(*next) )
    next++;

  if ( IS_STR_END_CHAR(*next) == *next )
    next= 0;
  
  return 1;
}

int _strconv(char const* str,float & val, char const* & next, bool must_have_end) {
  char* error_ptr;
  double tmp_double;

  if (0 == str) {
    next= 0;
    return 0;
  }

  tmp_double= strtod(str, &error_ptr);
  next= error_ptr;

  if ( str == next 
       || ERANGE == errno ) {
    next= 0;
    return 0;
  }

  if ( !must_have_end || IS_STR_END_CHAR(*next) || IS_WHITE_CHAR(*next) ) {
    val= tmp_double;
  }

  while ( IS_WHITE_CHAR(*next) )
    next++;

  if ( IS_STR_END_CHAR(*next) )
    next= 0;
  
  return 1;
}

int _strconv(char const* str,bool & val, char const* & next, bool must_have_end) {
  bool tmp_bool;

  if (0 == str) {
    next= 0;
    return 0;
  }

  while ( IS_WHITE_CHAR(*str) )
    str++;
  
  if ( '1' == *str ) {
    tmp_bool= true;
    next= str+1;
  }
  else if ( '0' == *str ) {
    tmp_bool= false;
    next= str+1;
  }
  else if (strncmp(str,"true",4)==0) {
    tmp_bool= true;
    next= str+4;
  }
  else if (strncmp(str,"false",5)==0) {
    tmp_bool= false;
    next= str+5;
  }
  else {
    next= 0;
    return 0;
  }

  if ( !must_have_end || IS_STR_END_CHAR(*next) || IS_WHITE_CHAR(*next) ) {
    val= tmp_bool;
  }

  while ( IS_WHITE_CHAR(*next) )
    next++;

  if ( IS_STR_END_CHAR(*next) )
    next= 0;
  
  return 1;
}

int strconv(char const* str,int & val, char const* & next) {
  return _strconv(str,val,next,true);
}
int strconv(char const* str,long & val, char const* & next) {
  return _strconv(str,val,next,true);
}
int strconv(char const* str,double & val, char const* & next) {
  return _strconv(str,val,next,true);
}
int strconv(char const* str,float & val, char const* & next) {
  return _strconv(str,val,next,true);
}
int strconv(char const* str,bool & val, char const* & next) {
  return _strconv(str,val,next,true);
}
/****************************************************************/
int substrconv(char const* str,int & val, char const* & next) {
  return _strconv(str,val,next,false);
}
int substrconv(char const* str,long & val, char const* & next) {
  return _strconv(str,val,next,false);
}
int substrconv(char const* str,double & val, char const* & next) {
  return _strconv(str,val,next,false);
}
int substrconv(char const* str,float & val, char const* & next) {
  return _strconv(str,val,next,false);
}
int substrconv(char const* str,bool & val, char const* & next) {
  return _strconv(str,val,next,false);
}

/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/
#if 0

#include <iostream>

int main(int argc, char **argv) {
  const char *chr, *dum;
  chr= argv[1];
  chr= "val             =         \t     3 4 \n 5";
  if (strshift(chr,"val",dum)) {
    chr= dum;
    cout << "\nafter val [" << chr << "]";
    if (strshift(chr,"=",dum)) {
      chr= dum;
      cout << "\nafter = [" << chr << "]";
    }
  }
  else {
    cerr << "\n not valid args";
  }
    
  while (chr) {
    int i= -1;
    strconv(chr,i,chr);
    cout << "\nnext= " << chr << flush;
    cout << "\ngot " << i << flush;
  }
  cout << endl;
  return 1;
}
#endif
