/*
 * Copyright (c) 1999 - 2001, Artur Merke <amerke@ira.uka.de> 
 *
 * This file is part of FrameView2d.
 *
 * FrameView2d is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * FrameView2d is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FrameView2d; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef _VECTOR2D_H_
#define _VECTOR2D_H_

#include <math.h>
#include <iostream>
#include "angle.h"

class Vector2d {
public:
   double   x ;
   double   y ;

   Vector2d(double vx = 0.0,double vy = 0.0) { x= vx; y= vy;}
   void init_polar(double n, const Angle& a);

   void    operator +=(const Vector2d& v) ;
   void    operator -=(const Vector2d& v) ;
   void    operator *=(double a) ;
   void    operator /=(double a) ;

   double norm() const { return sqrt(square(x)+square(y)) ;}
   
   /** berechnet das Argument der komplexen Zahl (x,y) auf das Intervall
      [0,2*PI) normalisiert. Das Argument von (0,0) wird auf 0 gesetzt*/
   Angle arg() const;
   
   friend Vector2d operator *(double a, const Vector2d& b);
   friend Vector2d operator +(const Vector2d& a, const Vector2d& b);
   friend Vector2d operator -(const Vector2d& a, const Vector2d& b);

   /** */
   //void normalize(double l = 1.0) ;

   /** gibt den euklidischen Abstand zwischen den Vektoren */
   double distance(const Vector2d& orig) const;
    
   /** gibt den Winkel zwischen dem Vector2d selbst und dem Vector2d dir */
   Angle angle(const Vector2d& dir) const;

   /** rotiert um den angegebenen Winkel */
   void rotate(const Angle& ang) ;
private:
   double square(double v) const { return v * v; }
};


extern ostream& operator<< (ostream& o, const Vector2d& v) ;


#endif
